/* version minerLC */

/*
 * directed_graph.hpp
 *
 *  Created on: 9 sept. 2015
 *      Author: Dominique Bouthinon
 *
 *  Representation d'un graphe oriente par une liste d'adjacences
 *  les valeurs des noeuds sont de n'importe quel type
 *  represente par le type template Tvalue
 *
 *  - ajout degres ponderes : Aout 2019
 *
 */

#ifndef DIRECTED_GRAPH_HPP_
#define DIRECTED_GRAPH_HPP_

#include <functional>
#include <iostream>
#include <algorithm>
#include <exception>
#include <stack>

#include "graph.hpp"
#include "types_graph.hpp"
#include "abstraction.hpp"
#include "utils_divers.hpp"

using namespace std ;

/* =============================== class graph ============================== */

template <typename Tvalue>
class directed_graph : public virtual graph<Tvalue>
{
	// ================================== PRIVATE ==============================

	private :

	void compute_hubs_autorities(float min_hub, float min_autority,
								 Tvertice_identifiers_set& hubs,
								 Tvertice_identifiers_set& autorities) ;

	void initialise_hubs_autorities(float min_hub, float min_autority,
									Tvertice_identifiers_set& hubs,
									Tvertice_identifiers_set& autorities,
									map<Tvertice_identifier, float>& degres_hubs,
									map<Tvertice_identifier, float>& degres_autorities) ;

			// retourne le nombre de laisons depuis A vers x
	virtual  float dout(Tvertice_identifier x, Tvertice_identifiers_set& A) ;

			// retourne le nombre de laisons de x vers H
	virtual float din(Tvertice_identifier x, Tvertice_identifiers_set& H) ;


			// verifie le point fixe de l'abstraction hub_aut
	bool point_fixe(map<Tvertice_identifier, float>& degres,
			   	   	map<Tvertice_identifier, float>& degres1) ;

			// abstraction hub_autorites : elimine les hubs ou autorites
			// non valide
	void elimine_noeuds_degre_zero(map<Tvertice_identifier, float>& degres) ;


/* ================================ PUBLIC ================================== */

	public:

	// --------------------------- CONSTRUCTEURS -------------------------------


	directed_graph() ;				// graph sans information

			 	 	 	 	// construit le graphe avec un nom et une fonction d'affichage
	directed_graph(
					string aName,
					function<void (Tvalue)> display_vertice_value_function
		 	 	  ) ;


					 // construit le graphe a partir d'une liste de noeuds et
					 // d'adjacences (avec predecesseurs et successeurs).
					 // ATTENTION : les objets internes sont des copies des parama�tres
	directed_graph(
					string aName,
					Tvertices_map<Tvalue>& vertices,
					Tadjacences_map& adjacences,
					function<void (Tvalue)> display_vertice_value_function
				   ) ;


	directed_graph(
					string aName,
					Tvertices_map<Tvalue>& vertices,
					Tadjacences_map& adjacences,
					function<void (Tvalue)> display_vertice_value_function,
					function<string (Tvalue)> toString_value_function
					) ;


	directed_graph(
					string aName,
					Tvertices_map<Tvalue>& vertices,
					function<void (Tvalue)> display_vertice_value_function
		 	 	 ) ;


					// constructeur par copie
	directed_graph(const directed_graph<Tvalue>& g) ;

	virtual ~directed_graph() ;

	graph<Tvalue>* copy_graph() ;


	// -------------------------- NOEUD ----------------------------------------

	// retourne l'ensemble des noeuds connectes au noeud ident
	void adjacent_vertices_out(Tvertice_identifier ident, Tvertice_identifiers_set& vertices) ;

	// retourne l'ensemble des noeuds connectes au noeud ident
	void adjacent_vertices_in(Tvertice_identifier ident, Tvertice_identifiers_set& vertices) ;


	/**
		 * retourne le degre entrant du noeud vertice (s'il existe)
		 * sinon retoune -1
		 */

	virtual float degree_in(Tvertice_identifier ident) ;

	/**
		 * retourne le degre sortant du noeud vertice (s'il existe)
		 * sinon retoune -1
		 */

	virtual float degree_out(Tvertice_identifier ident) ;


	/**
	 *   pour le cas d'un graphe oriente dont le graphe non oriente
	 *   associe serait multiple, alors qu'il faut le traiter en graphe simple.
	 *   retourne # degre_in + #degre_out - #(degre_in inter degree_out)
	*/

	float degre_simple(Tvertice_identifier ident) ;


	// ---------------------------- SAUVEGARDE -------------------------------

	// sauvegarde le graph en format dot
	void save_graph_dot(string name, bool vertice_ident, bool vertice_label) ;


	// ------------------------------ DISTANCE ---------------------------------

	// retourne D(x,k) l'ensemble des sommets se trouvant a distance_in k
	// du sommet x

	void distance_in(
	 			 Tvertice_identifier x,
				 int k,
				 Tvertice_identifiers_set& vertices,
				 Tvertice_identifiers_set& dk1,		// noeuds a distance out k-1
				 Tvertice_identifiers_set& dmaxk2	// noeuds a distance out au plus k-2
			    ) ;

		// retourne l'ensemble des sommets se trouvant a distance
		// au plus n du sommet x
	void distance_in_max(
					 Tvertice_identifier x,
					 int n,
					 Tvertice_identifiers_set& vertices
					) ;


		// retourne D(x,k) l'ensemble des sommets se trouvant a distance_out k
		// du sommet x
	void distance_out(
	 			 Tvertice_identifier x,
				 int k,
				 Tvertice_identifiers_set& vertices,
				 Tvertice_identifiers_set& dk1,		// noeuds a distance out k-1
				 Tvertice_identifiers_set& dmaxk2	// noeuds a distance out au plus k-2
			    ) ;


		// retourne l'ensemble des sommets se trouvant a distance
		// au plus n du sommet x
	void distance_out_max(
					 	 Tvertice_identifier x,
						 int n,
						 Tvertice_identifiers_set& vertices
						) ;

	// ----------------------------- ABSTRACTIONS ------------------------------

	bool abstraction_classic_hub_autority_pred(int min_sup, float min_hub,
											   float min_autority,
											   Tvertice_identifiers_set& H,
											   Tvertice_identifiers_set& A) ;

	bool abstraction_classic_hub_autority(int min_sup, float min_hub, float min_autority) ;

	// bool abstraction_hub_autority(int min_sup, int hub, int autority) ;

				// redefinition
	bool abstraction_minimal_degree_in_out(int min_sup, float min_in, float min_out) ;

		/*
		 * on retire tous les noeuds (et arcs y faisant reference) qui ne sont
		 * pas de degre SIMPLE min_degre
		 * SI LES COMPOSANTES CONNEXES ONT ETE CALCULEES,
		 * ELLES SONT MISES aJOUR EN CONSEQUENCES
		 */
	bool abstraction_minimal_degre_simple(int min_sup, float min_degre) ;

	bool abstraction_near_star_in(float min_degre_in, int max_distance) ;

	bool abstraction_near_star_out(float min_degre_out, int max_distance) ;
 
} ;
// fin classe directed_graph<Tvalue>

/* ===================== IMPLEMENTATIONS FONCTIONS ========================== */

//------------------------------------ CONSTRUCTEURS ---------------------------

template <typename Tvalue>
directed_graph<Tvalue>::
directed_graph() : graph<Tvalue>()
{

}

		 	 	 	 	// construit le graphe avec un nom et une fonction d'affichage
template <typename Tvalue>
directed_graph<Tvalue>::
directed_graph(
				string aName,
				function<void (Tvalue)> display_vertice_value_function
	 	 	 ) :
			 graph<Tvalue>(aName, display_vertice_value_function)
{
}
				 // construit le graphe a partir d'une liste de noeuds et
				 // d'adjacences (avec predecesseurs et successeurs).
				 // ATTENTION : les objets internes sont des copies des parama�tres

template <typename Tvalue>
directed_graph<Tvalue>::
directed_graph(
				string aName,
				Tvertices_map<Tvalue>& vertices,
				Tadjacences_map& adjacences,
				function<void (Tvalue)> display_vertice_value_function
			   ) : graph<Tvalue>(aName, vertices, adjacences, display_vertice_value_function)
{
}

// construit le graphe a partir d'une liste de noeuds et
// d'adjacences (avec predecesseurs et successeurs).
// ATTENTION : les objets internes sont des copies des paramatres
template <typename Tvalue>
directed_graph<Tvalue>::
directed_graph(
				string aName,
				Tvertices_map<Tvalue>& vertices,
				Tadjacences_map& adjacences,
				function<void (Tvalue)> display_vertice_value_function,
				function<string (Tvalue)> toString_value_function
				) : graph<Tvalue>(aName, vertices, adjacences,
						          display_vertice_value_function,
								  toString_value_function)
{
}


template <typename Tvalue>
directed_graph<Tvalue>::
directed_graph(
				string aName,
				Tvertices_map<Tvalue>& vertices,
				function<void (Tvalue)> display_vertice_value_function
	 	 	 ) : graph<Tvalue>(aName, vertices, display_vertice_value_function)
{

}

				// constructeur par copie
template <typename Tvalue>
directed_graph<Tvalue>::
directed_graph(const directed_graph<Tvalue>& g) : graph<Tvalue>(g)
{
}

				// DESTRUCTEUR
template <typename Tvalue>
directed_graph<Tvalue>::
~directed_graph()
{
}


// ----------------------------- COPIE DU GRAPHE ---------------------------


template <typename Tvalue>
graph<Tvalue>* directed_graph<Tvalue>::copy_graph()
{
	return new directed_graph<Tvalue>(*this) ;
}


//---------------------------- NOEUD----------------------------------------

// retourne l'ensemble des noeuds connectes au noeud ident
template <typename Tvalue>
void directed_graph<Tvalue>::
adjacent_vertices_out(Tvertice_identifier ident, Tvertice_identifiers_set& vertices)
{
	vertices.insert(this->get_successors(ident).begin(),
					this->get_successors(ident).end()) ;
}


// retourne l'ensemble des noeuds connectes au noeud ident
template <typename Tvalue>
void directed_graph<Tvalue>::
adjacent_vertices_in(Tvertice_identifier ident, Tvertice_identifiers_set& vertices)
{
	vertices.insert(this->get_predeccessors(ident).begin(),
					this->get_predeccessors(ident).end()) ;
}

	/**
	 * retourne le degre entrant du noeud vertice (s'il existe)
	 * sinon retoune -1
	 */
	template <typename Tvalue>
	float directed_graph<Tvalue>::degree_in(Tvertice_identifier ident)
	{
		if (!this->in_graph(ident))
			return -1 ;

		return this->get_predeccessors(ident).size() ;
	}


	/**
	 * retourne le degre sortant du noeud vertice (s'il existe)
	 * sinon retoune -1
	 */
	template <typename Tvalue>
	float directed_graph<Tvalue>::degree_out(Tvertice_identifier ident)
	{
		if (!this->in_graph(ident))
			return -1 ;

		return this->get_successors(ident).size() ;
	}

	/**
		 *   pour le cas d'un graphe oriente dont le graphe non oriente
		 *   associe serait multiple, alors qu'il faut le traiter en graphe simple.
		 *   retourne # degre_in + #degre_out - #(degre_in inter degree_out)
		*/
	template <typename Tvalue>
	float directed_graph<Tvalue>::degre_simple(Tvertice_identifier ident)
	{
			if (!this->in_graph(ident))
				return -1 ;

			Tvertice_identifiers_set inter ;

			set_intersection (
					           this->get_predeccessors(ident).begin(), this->get_predeccessors(ident).end(),
							   this->get_successors(ident).begin(), this->get_successors(ident).end(),
					           inserter(inter,inter.begin())
							  ) ;

			return  this->get_predeccessors(ident).size() +
					this->get_successors(ident).size()   -
				    inter.size() ;
	}


	// -------------------------- DISTANCE -------------------------------------


	// retourne D(x,k) l'ensemble des sommets se trouvant a distance_in k
	// du sommet x
	// on utilise les proprietes suivantes :
	// D(x, 0) = {x}
	// D(x, 1) = {y | y directement connecte a x}
	// et pour k >=2 :
	// D(x, k) = {y | il existe z dans D(x,k-1) tel que
    //                y connecte a z et y n'est pas dans Dmax(x, k-2)

template <typename Tvalue>
void directed_graph<Tvalue>::
distance_in(
 			 Tvertice_identifier x,
			 int k,
			 Tvertice_identifiers_set& vertices,
			 Tvertice_identifiers_set& dk1,		// noeuds a distance out k-1
			 Tvertice_identifiers_set& dmaxk2	// noeuds a distance out au plus k-2
		    )
{
	if (k==0)
	{
		vertices.insert(x) ;
		return ;
	}

	if (k==1)
	{
		this->adjacent_vertices_in(x, vertices) ;
		return ;
	}
									// k >= 2

	Tvertice_identifiers_set_iterator it_z     = dk1.begin() ;
	Tvertice_identifiers_set_iterator it_z_end = dk1.end() ;

									// pour chaque noeud z a distance out k-1 de x
	for ( ; it_z != it_z_end ; it_z++)
	{
									// chaque noeud y connecte a z (pas deja traite)
									// qui n'est pas a une distance inferieure a k de x
									// est a distance k de x

									// on explore les predecesseurs de z
		Tvertice_identifiers_set_iterator it_y     = this->get_predeccessors(*it_z).begin() ;
		Tvertice_identifiers_set_iterator it_y_end = this->get_predeccessors(*it_z).end() ;

		for ( ; it_y != it_y_end ; it_y++)
		{
									// on ajoute le noeud y s'il n'est pas
									// dans dmaxk2
									// s'il est deja dans vertices
									// l'insertion ne se fera pas (set)
			if (dmaxk2.find(*it_y) == dmaxk2.end())
			{
				vertices.insert(*it_y) ;
			}
		}
	}
}

	// retourne l'ensemble des sommets se trouvant a distance
	// au plus n du sommet x
	// on utilise les proprietes suivantes :
	// Dout_max(x, 0) = {x}
	// pour k >= 1
	// Dout_max(x, k) = Dout_max(x, k-1) union D_out(x, k)
	// (D_out(x, k) defini dans le fonction precedente)
	// on utilise ici une fonction iterative (non recursive)

template <typename Tvalue>
void directed_graph<Tvalue>::
distance_in_max(
				 Tvertice_identifier x,
				 int n,
				 Tvertice_identifiers_set& vertices
				)
{
	vertices.insert(x) ;					// vertices contient le noeud
											// a distance 0 de x
	if (n == 0)
		return ;

	Tvertice_identifiers_set* dk1 = new Tvertice_identifiers_set() ;
											// noeuds a distance 1 de x
											// (et a distance k-1) plus loin
	this->adjacent_vertices_in(x, *dk1) ;

	vertices.insert(dk1->begin(), dk1->end()) ;

											// vertices contient les noeuds
											// a distance au plus 1 de x
	if (n==1)
			return ;
											// n >= 2
											// on iterera pour k = 2 a n

	Tvertice_identifiers_set dmaxk2 ;		// noeuds a distance au plus
	dmaxk2.insert(x) ;						// k-2 de x

											// pour chaque distance k
											// de 2 a n
	for (int k = 2 ; k <= n ; k++)
	{
		Tvertice_identifiers_set* dk = new Tvertice_identifiers_set() ;

											// recupere dans dk les noeuds
											// a distance k de x
		this->distance_in(x, k, *dk, *dk1, dmaxk2) ;

											// les noeuds a distance au plus k
											// sont dans dmaxk2 union dk1 union dk
		if (k < n)
		{
			dmaxk2.insert(dk1->begin(), dk1->end()) ;

			delete dk1 ;
			dk1 = dk   ;
		}
		else
		{
											// les noeuds a distance au plus n
											// sont dmaxk2 union dk1 union dk
			vertices.insert(dmaxk2.begin(), dmaxk2.end()) ;
			vertices.insert(dk1->begin(), dk1->end()) ;
			vertices.insert(dk->begin(), dk->end()) ;

			delete dk1 ;
			delete dk ;
		}
	}
}

	// retourne D(x,k) l'ensemble des sommets se trouvant a distance_out k
	// du sommet x
	// on utilise les proprietes suivantes :
	// D(x, 0) = {x}
	// D(x, 1) = {y | y directement connecte a x}
	// et pour k >=2 :
	// D(x, k) = {y | il existe z dans D(x,k-1) tel que
    //                y connecte a z et y n'est pas dans Dmax(x, k-2)

template <typename Tvalue>
void directed_graph<Tvalue>::
distance_out(
 			 Tvertice_identifier x,
			 int k,
			 Tvertice_identifiers_set& vertices,
			 Tvertice_identifiers_set& dk1,		// noeuds a distance out k-1
			 Tvertice_identifiers_set& dmaxk2	// noeuds a distance out au plus k-2
		    )
{
	if (k==0)
	{
		vertices.insert(x) ;
		return ;
	}

	if (k==1)
	{
		this->adjacent_vertices_out(x, vertices) ;
		return ;
	}
									// k >= 2

	Tvertice_identifiers_set_iterator it_z     = dk1.begin() ;
	Tvertice_identifiers_set_iterator it_z_end = dk1.end() ;

									// pour chaque noeud z a distance out k-1 de x
	for ( ; it_z != it_z_end ; it_z++)
	{
									// chaque noeud y connecte a z (pas deja traite)
									// qui n'est pas a une distance inferieure a k de x
									// est a distance k de x

									// on explore les successeurs de z
		Tvertice_identifiers_set_iterator it_y     = this->get_successors(*it_z).begin() ;
		Tvertice_identifiers_set_iterator it_y_end = this->get_successors(*it_z).end() ;

		for ( ; it_y != it_y_end ; it_y++)
		{
									// on ajoute le noeud y s'il n'est pas
									// dans dmaxk2
									// s'il est deja dans vertices
									// l'insertion ne se fera pas (set)
			if (dmaxk2.find(*it_y) == dmaxk2.end())
			{
				vertices.insert(*it_y) ;
			}
		}
	}
}

	// retourne l'ensemble des sommets se trouvant a distance
	// au plus n du sommet x
	// on utilise les proprietes suivantes :
	// Dout_max(x, 0) = {x}
	// pour k >= 1
	// Dout_max(x, k) = Dout_max(x, k-1) union D_out(x, k)
	// (D_out(x, k) defini dans le fonction precedente)
	// on utilise ici une fonction iterative (non recursive)

template <typename Tvalue>
void directed_graph<Tvalue>::
distance_out_max(
				 Tvertice_identifier x,
				 int n,
				 Tvertice_identifiers_set& vertices
				)
{
	vertices.insert(x) ;					// vertices contient le noeud
											// a distance 0 de x
	if (n == 0)
		return ;

	// cout << "x = " << x << endl ;


	Tvertice_identifiers_set* dk1 = new Tvertice_identifiers_set() ;
											// noeuds a distance 1 de x
											// (et a distance k-1) plus loin
	this->adjacent_vertices_out(x, *dk1) ;

	vertices.insert(dk1->begin(), dk1->end()) ;

											// vertices contient les noeuds
											// a distance au plus 1 de x
	if (n==1)
			return ;
											// n >= 2
											// on iterera pour k = 2 a n

	Tvertice_identifiers_set dmaxk2 ;		// noeuds a distance au plus
	dmaxk2.insert(x) ;						// k-2 de x

											// pour chaque distance k
											// de 2 a n
	for (int k = 2 ; k <= n ; k++)
	{
		Tvertice_identifiers_set* dk = new Tvertice_identifiers_set() ;

											// recupere dans dk les noeuds
											// a distance k de x
		this->distance_out(x, k, *dk, *dk1, dmaxk2) ;

											// les noeuds a distance au plus k
											// sont dans dmaxk2 union dk1 union dk
		if (k < n)
		{
			dmaxk2.insert(dk1->begin(), dk1->end()) ;

			delete dk1 ;
			dk1 = dk   ;
		}
		else
		{
											// les noeuds a distance au plus n
											// sont dmaxk2 union dk1 union dk
			vertices.insert(dmaxk2.begin(), dmaxk2.end()) ;
			vertices.insert(dk1->begin(), dk1->end()) ;
			vertices.insert(dk->begin(), dk->end()) ;

			delete dk1 ;
			delete dk ;
		}
	}
}

	// --------------------------- ABSTRACTIONS --------------------------------

	// conserve un sommet s'il est de degre superieur ou egale a min_degre,
	// ou s'il est a distance au plus max_distance d'un sommet de degre
	// superieur ou egal a min_degre in


// retourne le nombre de laisons de x vers A
template <typename Tvalue>
float directed_graph<Tvalue>::dout(Tvertice_identifier x, Tvertice_identifiers_set& A)
{
	Tvertice_identifiers_set succ ;
	Tvertice_identifiers_set to_A ;

	this->adjacent_vertices_out(x, succ) ;	// noeuds successeurs de x

	setIntersection(&to_A, succ, A); 		// noeuds de X successeurs de x

	return to_A.size() ;
}


	// retourne le nombre de laisons de H vers x
template <typename Tvalue>
float directed_graph<Tvalue>::din(Tvertice_identifier x, Tvertice_identifiers_set& H)
{
	Tvertice_identifiers_set pred   ;
	Tvertice_identifiers_set from_H ;

		this->adjacent_vertices_in(x, pred) ;	// noeuds predecesseurs de x

		setIntersection(&from_H, pred, H);  	// noeuds de X predecesseurs de x

		return from_H.size() ;
}



// calcul des sommets dont le degre_out
// est superieur a h
template <typename Tvalue>
void directed_graph<Tvalue>::compute_hubs_autorities(float min_hub, float min_autority,
										   	   	   	 Tvertice_identifiers_set& hubs,
										             Tvertice_identifiers_set& autorities)
{
	Tvertice_identifiers_set_iterator it = this->get_vertice_identifiers_list().begin();

											// pour chaque sommet du graphe
	while(it != this->get_vertice_identifiers_list().end())
	{
		// cout << "*it = " << *it << endl ;

		float degree_out = this->degree_out(*it) ;

		if (degree_out >= min_hub)
		{
			hubs.insert(*it) ;
		}

		float degree_in = this->degree_in(*it) ;

		if (degree_in >= min_autority)
		{
			autorities.insert(*it) ;
		}

		it++ ;
	}
}




// calcul des sommets dont le degre_out
// est superieur a h
template <typename Tvalue>
void directed_graph<Tvalue>::initialise_hubs_autorities(float min_hub, float min_autority,
										   Tvertice_identifiers_set& hubs,
										   Tvertice_identifiers_set& autorities,
									   	   map<Tvertice_identifier, float>& degres_hubs,
										   map<Tvertice_identifier, float>& degres_autorities)
{
	Tvertice_identifiers_set_iterator it = this->get_vertice_identifiers_list().begin();

											// pour chaque sommet du graphe
	while(it != this->get_vertice_identifiers_list().end())
	{
		// cout << "*it = " << *it << endl ;

		float degree_out = this->degree_out(*it) ;

		if (degree_out >= min_hub)
		{
			hubs.insert(*it) ;
			degres_hubs.insert(make_pair(*it, degree_out)) ;
		}

		float degree_in = this->degree_int(*it) ;

		if (degree_in >= min_autority)
		{
			autorities.insert(*it) ;
			degres_autorities.insert(make_pair(*it, degree_in)) ;
		}

		it++ ;
	}
}

	// point fixe atteint si pour chaque element de degres1 est
	// egal a� son homologue dans degres
	// on ne tient pas compte

template <typename Tvalue>
bool directed_graph<Tvalue>::point_fixe(map<Tvertice_identifier, float>& degres,
		   	   	   	   	   	   	   	    map<Tvertice_identifier, float>& degres1)
{
	bool point_fixe = true ;

	map<Tvertice_identifier,float>::iterator d  = degres.begin() ;
	map<Tvertice_identifier,float>::iterator d1 = degres1.begin() ;

										// pour chaque NOUVEAU hub on calcule son degre en fonction
										// des autorites precedentes
	while (d != degres.end() && point_fixe)
	{
		if (d1-> second != 0)			// seuls les noeuds encore valides comptent pour le point fixe
		{
			point_fixe = (d->second == d1->second) ;
		}

		if (point_fixe)
		{
			d++ ;
			d1++ ;
		}
	}

	return point_fixe ;
}


			// elimine les noeuds de degre zero
template <typename Tvalue>
void directed_graph<Tvalue>::elimine_noeuds_degre_zero(map<Tvertice_identifier, float>& degres)
{

	map<Tvertice_identifier,float>::iterator d  = degres.begin() ;

										// pour chaque NOUVEAU hub on calcule son degre en fonction
										// des autorites precedentes
	while (d != degres.end())
	{
		if (d->second == 0)
		{
			d = degres.erase(d) ;
		}
		else
			d++ ;
	}
}


template <typename Tvalue>
bool directed_graph<Tvalue>::abstraction_classic_hub_autority_pred(
																	int min_sup,
																	float min_hub,
																	float min_autority,
																	Tvertice_identifiers_set& H,
																	Tvertice_identifiers_set& A
																   )
	{
		bool discarded_vertice = false ;

		Tvertice_identifiers_set Hp ;	// hubs iteration precedente
		Tvertice_identifiers_set Ap ;	// autorites iteration precedente

		H.clear() ;						// important : retire les hubs des essais precedents
		A.clear() ;						// idem

										// calcul des hubs et autorites
		compute_hubs_autorities(min_hub, min_autority, H, A) ;

		do
		{
			Hp = H ;			// hubs iteration precedente
			Ap = A ;			// autorites iteration precedente

			// displaySet("Hp", Hp) ;
			// cout << endl ;

			// displaySet("Ap", Ap) ;
			// cout << endl ;

			H.clear() ;
			A.clear() ;

			Tvertice_identifiers_set_iterator x = this->get_vertice_identifiers_list().begin() ;

														// parcours des sommets
			while (this->nb_vertices() >= min_sup && x != this->get_vertice_identifiers_list().end())
			{

				float dH = this->dout(*x, Ap) ;
				float dA = this->din(*x, Hp) ;

				// cout << "*x = " << *x << endl ;
				// cout << "dH = " << dH << endl ;
				// cout << "dA = " << dA << endl ;

				if (dH < min_hub && dA < min_autority)	// x ni hub ni autorite
				{
					Tvertice_identifiers_set out_autorities ;
					Tvertice_identifiers_set adja_out ;

					// sauf dans vertices_list
					this->discard_vertice_elements(*x, true, true, false) ;

													// on supprime (*x) de vertices_list
					x = this->get_vertice_identifiers_list().erase(x) ;

					discarded_vertice = true ;
													// pointFixe = false ;  // modif demande par Henry
				  }
				  else
				  {
					  if (dH >= min_hub)
					  {
						  H.insert(*x) ;
						  // cout << "Hub" << endl ;
					  }

					  if (dA >= min_autority)
					  {
						  A.insert(*x) ;
						  // cout << "Aut" << endl ;
					  }
						x++ ;
				  }
			}
			// fin while parcours sommets

			// displaySet("H", H) ;
			// cout << endl ;

			// displaySet("A", A) ;
			// cout << endl ;

			// cout << "A.size  = " << A.size()  << endl ;
			// cout << "Ap.size = " << Ap.size() << endl ;

			// cout << "H.size  = " << H.size()  << endl ;
			// cout << "Hp.size = " << Hp.size() << endl ;
		}
		while ((A.size() != Ap.size()) || (H.size() != Hp.size())) ;

		return discarded_vertice ;
	}
	// fin abstraction_classic_hub_autority_pred



	// provisoirement
template <typename Tvalue>
bool directed_graph<Tvalue>::abstraction_classic_hub_autority(int min_sup,
															  float min_hub,
															  float min_autorite)
{
	return false ;
}



template <typename Tvalue>
bool directed_graph<Tvalue>::abstraction_near_star_in(float min_degree_in, int max_distance)
{
	Tvertice_identifiers_set vertices_ok ;
	Tvertice_identifiers_set_iterator it = this->get_vertice_identifiers_list().begin();

											// pour chaque sommet du graphe
	while(it != this->get_vertice_identifiers_list().end())
	{
		// cout << "*it = " << *it << endl ;

		float degree_in = this->degree_in(*it) ;

		// cout << "degree = " << degree << endl ;

		if (degree_in >= min_degree_in)
		{
			Tvertice_identifiers_set dmaxk ;

			this->distance_in_max(*it, max_distance, dmaxk) ;

			// displaySet("dmaxk", dmaxk) ;

			vertices_ok.insert(dmaxk.begin(), dmaxk.end()) ;
		}

		it++ ;
	}

	if (vertices_ok.size() < this->nb_vertices())
	{
											// retire les sommets qui ne sont pas ok
		Tvertice_identifiers_set diff ;

		set_difference(
						this->get_vertice_identifiers_list().begin(),
						this->get_vertice_identifiers_list().end(),
						vertices_ok.begin(), vertices_ok.end(),
						inserter(diff, diff.begin())
					   ) ;

		// graph::display_vertice_identifiers_set(diff) ;
		// cout << endl ;

		// displaySet("diff", diff) ;
											 // diff contient les noeuds a� retirer
		this->discard_vertices(diff) ;

		return true ;
	}

	return false ;							// aucun sommet retire
}



	// conserve un sommet s'il est de degre superieur ou egale a� min_degre,
	// ou s'il est a� distance au plus max_distance d'un sommet de degre
	// superieur ou egal a� min_degre out
template <typename Tvalue>
bool directed_graph<Tvalue>::abstraction_near_star_out(float min_degree_out, int max_distance)
{
	Tvertice_identifiers_set vertices_ok ;
	Tvertice_identifiers_set_iterator it = this->get_vertice_identifiers_list().begin() ;

											// pour chaque sommet du graphe
	while(it != this->get_vertice_identifiers_list().end())
	{
		// cout << "*it = " << *it << endl ;

		float degree_out = this->degree_out(*it) ;

		// cout << "degree_out = " << degree_out << endl ;

		if (degree_out >= min_degree_out)
		{
			Tvertice_identifiers_set dmaxk ;

			this->distance_out_max(*it, max_distance, dmaxk) ;

			vertices_ok.insert(dmaxk.begin(), dmaxk.end()) ;

			// graph<Tvalue>::display_vertice_identifiers_set(vertices_ok) ;
		}

		it++ ;
	}

	// cout << "fin max distance out " << endl << endl ;


	if (vertices_ok.size() < this->nb_vertices())
	{
											// retire les sommets qui ne sont pas ok
		Tvertice_identifiers_set diff ;

		set_difference(
						this->get_vertice_identifiers_list().begin(),
						this->get_vertice_identifiers_list().end(),
						vertices_ok.begin(), vertices_ok.end(),
						inserter(diff, diff.begin())
					   ) ;

		// graph::display_vertice_identifiers_set(diff) ;
		// cout << endl ;

		// displaySet("diff", diff) ;
			  	  	  	  	  	  	  	  	 // diff contient les noeuds a� retirer
		this->discard_vertices(diff) ;

		return true ;
	}

	return false ;							// aucun sommet retire
}



	//			voir si on peut utiliser les lambdas ou pointeur de fonction
	//			pour avoir une methode generique (valable pour plusieurs fonction
	/*
	 * on retire tous les noeuds (et arcs y faisant reference) qui ne sont
	 * pas de degre entrant au moins in, et de degre sortant au moint out
	 * SI LES COMPOSANTES CONNEXES ONT ETE CALCULEES,
	 * ELLES SONT MISES aJOUR EN CONSEQUENCES
	 */
	template <typename Tvalue>
	bool directed_graph<Tvalue>::abstraction_minimal_degree_in_out(int min_sup, float min_in, float min_out)
	{
		bool discarded_vertice = false ;
		bool changed = true ;

		while(this->nb_vertices() >= min_sup && changed)
		{
			changed = false ;

			Tvertice_identifiers_set_iterator it = this->get_vertice_identifiers_list().begin() ;

			while (this->nb_vertices() >= min_sup && it != this->get_vertice_identifiers_list().end())
			{
				if (this->degree_in(*it) < min_in || this->degree_out(*it) < min_out)
				{
					changed = true ;
													// on supprime (*it) dans toutes les donnees oa� il apparait
													// sauf dans vertices_list
					this->discard_vertice_elements(*it, true, true, false) ;

													// on supprime (*it) de vertices_list
					it = this->get_vertice_identifiers_list().erase(it) ;

					discarded_vertice = true ;
				}
				else
					it++ ;
			}
		}

		return discarded_vertice ;
	}

		/*
		 * on retire tous les noeuds (et arcs y faisant reference) qui ne sont
		 * pas de degre SIMPLE min_degre
		 * SI LES COMPOSANTES CONNEXES ONT ETE CALCULEES,
		 * ELLES SONT MISES A JOUR EN CONSEQUENCES
		 */
		template <typename Tvalue>
		bool directed_graph<Tvalue>::abstraction_minimal_degre_simple(int min_sup, float min_degre)
		{
			bool discarded_vertice = false ;
			bool changed = true ;

			while(this->nb_vertices() >= min_sup && changed)
			{
				changed = false ;

				Tvertice_identifiers_set_iterator it = this->get_vertice_identifiers_list().begin() ;

				while (this->nb_vertices() >= min_sup && it != this->get_vertice_identifiers_list().end())
				{
					if (this->degre_simple(*it) < min_degre)
					{
						changed = true ;
												// on supprime (*it) dans toutes les donnees ou il apparait
												// sauf dans vertices_list
						this->discard_vertice_elements(*it, true, true, false) ;

														// on supprime (*it) de vertices_list
						it = this->get_vertice_identifiers_list().erase(it) ;

						discarded_vertice = true ;
					}
					else
						it++ ;
				}
			}

			return discarded_vertice ;
		}


/* ============================== SAUVEGARDE GRAPHE ========================= */

// sauvegarde le graph en format dot
template <typename Tvalue>
void directed_graph<Tvalue>::save_graph_dot(string name, bool vertice_ident,
										    bool vertice_label)
{
	this->save_graph_dot2(name, vertice_ident, vertice_label, true) ;
}


#endif /* DIRECTED_GRAPH_HPP_ */
