/* version minerLC */

/*
 * graph.hpp
 *
 *  Created on: 9 sept. 2015
 *      Author: Dominique Bouthinon
 *
 *  - 8 Octobre 2019 : copy_graph
 *
 *  Representation d'un graphe par une liste d'adjacences
 *  les valeurs des noeuds sont de n'importe quel type
 *  represente par le type template Tvalue
 *
 */

#ifndef GRAPH_HPP_
#define GRAPH_HPP_

#include <functional>
#include <iostream>
#include <algorithm>
#include <exception>
#include <stack>
#include <fstream>
#include <cstdlib>
#include <set>

#include "utils_divers.hpp"
#include "types_graph.hpp"
#include "abstraction.hpp"
#include "roles.hpp"
#include "statsModularity.hpp"

using namespace std ;

class edges_graph ;
class weigthed_graphe ;

/* ============================== class graph ============================== */

template <typename Tvalue>
class graph
{
	// ================================== PRIVATE ==============================

	private :

	string 					 name ;									// nom du graphe
	Tvertice_identifiers_set vertice_identifiers_set ;				// liste des identifiants des noeuds du graphe

	Tvertices_map<Tvalue>    vertices_map ;							// noeuds du graphe
	Tadjacences_map  	     adjacences_map ;						// adjacences de chaque noeud

	map<Tvertice_identifier,  float> map_degres ;	// map dont chaque element <ident, degre> contient
													// le degre du noeud d'identifiant ident
													// Cette map est calcule par compute_map_degres
													// N'EST PAS MIS A JOUR PAR LES ABSTRACTIONS
													// (degre est float pour le cas des aretes ponderees
					
	bool					   recompute_connected_components = false ;
	bool 		  			   connected_components_computed  = false ;	// true si les composantes connexes
																		// ont ete calculees
	int 		  		       next_connected_component_ident = 0 ;		// prochaine identifiant de composante connexe
																		// a attribuer
	Tconnected_components      connected_components ;

	// ------------------------------ CREATION DU GRAPH ------------------------

	void build_vertice_identifiers() ;

	//------------------------- FIND A VERTICE ---------------------------------

						// retourne un iterateur sur le vertice d'identifiant ident
						// dans le vertices_map
	Tvertices_map_iterator<Tvalue> vertice_map_iterator(Tvertice_identifier ident) ;

						// retourne un iterateur sur les adjacences
						// du vertice d'identifiant ident
	Tadjacences_map_iterator adjacences_map_iterator(Tvertice_identifier ident) ;


	// ----------------------------- GESTION NOEUD INTERNE ---------------------

	void discard_vertice_adjacences(Tvertice_identifier ident) ;


	// ------------------------------ COMPOSANTES CONNEXES----------------------

						// retourne les composantes connexes du graph
	void internal_connected_components(Tvertice_identifiers_set& remaining_vertices) ;

						// met -1 comme numero de composante connexe
						// associe a chaque noeud
	void reset_component_vertices() ;

	// ------------------------------ ABSTRACTIONS -----------------------------

						// retourne true si les noeuds x et y ont au moins k
						// voisins communs
	bool sufficient_common_adjacent_vertices(const Tvertice_identifier& x, const Tvertice_identifier& y, int k) ;

	// -------------------------------- CLIQUE ---------------------------------

	// retourne true s'il existe une k-clique dans vertices
bool check_k_clique(
	const int k,												// nombre de noeuds a ajouter a clique
	Tvertice_identifiers_set& vertices,							// noeuds pouvant etre ajoutes a clique
	Tvertice_identifiers_set& clique);							// clique en construction


/* ================================ PUBLIC ================================== */

	public:

	// --------------------------- CONSTRUCTEURS -------------------------------

	graph() ;				// graph sans information

			 	 	 	 	// construit le graphe avec un nom et une fonction d'affichage
	graph(
			string aName,
			function<void (Tvalue)> display_vertice_value_function,
			function<string (Tvalue)> toString_value_function
		 ) ;


					 // construit le graphe a partir d'une liste de noeuds et
					 // d'adjacences (avec predecesseurs et successeurs).
					 // ATTENTION : les objets internes sont des copies des paramatres
	graph(
			string aName,
			Tvertices_map<Tvalue>& vertices,
			Tadjacences_map& adjacences,
			function<void (Tvalue)> display_vertice_value_function,
			function<string (Tvalue)> toString_value_function
		  ) ;

	graph(
				string aName,
				Tvertices_map<Tvalue>& vertices,
				function<void (Tvalue)> display_vertice_value_function,
				function<string (Tvalue)> toString_value_function
		 ) ;


					// constructeur par copie
	graph(const graph<Tvalue>& g) ;

					// destructeur polymorphe
	virtual ~graph() ;

	virtual graph<Tvalue>* copy_graph() ;


	// ------------------------------------------------------------------------

	function<void (Tvalue)>    display_value  ;					// pour afficher la valeur d'un noeud

	function<string (Tvalue)>  toString_value  ;				// pour convertir la valeur d'un noeud en string


	//---------------------------- COPIE GRAPHE--------------------------------


	// ------------------------- SAUVEGARDE ET VISU ---------------------------

					// sauvegarde le graph en format dot (utilise par la fonction
					// save_graph_dot)
	virtual void save_graph_dot2(string name, bool vertice_ident, bool vertice_label,
								 bool directed_graph) ;

					// sauvegarde le graph en format dot
	virtual void save_graph_dot(string name, bool vertice_ident, bool vertice_label) ;

					// visualise le graph
	virtual void visu_graph_dot(string graph_name, bool vertice_ident,
								bool vertice_label) ;


	// ---------------------------- GETTERS -----------------------------------

	Tvertice_identifiers_set& get_vertice_identifiers_list() ;

	Tvertice_identifiers_set get_copy_vertice_identifiers_list() ;

	string get_graph_name() ;

					// retourne une reference sur le noeud interne d'identifiant ident
	Tvertice<Tvalue>& get_vertice(Tvertice_identifier ident) ;

					// retourne une reference sur la valeur associee
					// au noeud d'identifiant ident
	Tvalue& get_vertice_value(Tvertice_identifier ident) ;

	Tvalue& get_vertice_value(Tvertices_map_iterator<Tvalue> it)  ;

					// retourne une reference sur le numero de composante
					// associe au noeud d'identifiant ident
	int& get_vertice_component(Tvertice_identifier ident) ;

					// retourne une reference sur le numero de composante
					// associe au noeud pointe par it
	int& get_vertice_component(Tvertices_map_iterator<Tvalue> it)  ;

					// retourne une reference sur le label
					// associe au noeud d'identifiant ident
	string& get_vertice_label(Tvertice_identifier ident) ;

					// retourne une reference sur le label
					// associe au noeud pointe par it
	string& get_vertice_label(Tvertices_map_iterator<Tvalue> it)  ;

					// retourne une reference sur la liste d'adjacences du graphe
	Tadjacences_map& get_adjacences_map() ;

	 	 	 	 	 // retourne une reference sur la liste des noeuds du graphe
	Tvertices_map<Tvalue>& get_vertices_map() ;

	// ---------------------------- SETTERS ------------------------------------

	void set_vertices(Tvertices_map<Tvalue>& vertices_map) ;

	void set_adjacences(Tadjacences_map& adjacences_map) ;

	// ------------------------- INFORMATIONS SUR LE GRAPHE --------------------

	int nb_vertices() ;


	/* retourne le nombre d'aretes :
	 * (ATTENTION : en interne le graphe contient l'arc (a-b) et l'arc (b-a)
	 *  mais on ne contera qu'un arc)
	 */
	int nb_edges() ;


		// retourne un map dont chaque element <ident, degre>
		// contient le degre du noeud d'identifiant ident
	void compute_map_degres() ;

		// retourne le degre initial du noeud d'identifiant iden
	float get_map_degre_vertice(Tvertice_identifier ident) ;

		// retourne la map des degres initiaux des noeuds du graphe
	map<Tvertice_identifier,  float>&  get_map_degres() ;

	// -------------------------- AFFICHAGE ------------------------------------

										// graph
	virtual void display_graph() ;
	virtual void display_graph(string message) ;

										// vertice
	void display_vertice_value(Tvertice_identifier ident) ;
	void display_vertice_value_it(Tvertices_map_iterator<Tvalue> it) ;
	void display_vertice(Tvertice_identifier ident) ;
	void display_vertice_it(Tvertices_map_iterator<Tvalue> it) ;

										// set of vertices
	void display_vertices() ;
	void display_vertices_values() ;
	void display_vertice_identifiers() ;
	static void display_vertice_identifiers_set(Tvertice_identifiers_set idents_set) ;
	static string toString_vertice_identifiers_set(Tvertice_identifiers_set idents_set) ;
	void display_vertices_set() ;
	void display_vertices_set(const Tvertice_identifiers_set& idents_set) ;
	void display_vertices_idents_values_set(Tvertice_identifiers_set& vertices) ;
	// void display_vertices_sets(set<Tvertice_identifiers_set>& idents_sets) ;

										// edges
	void display_edges() ;

										// display connected components
	void display_connected_components() ;
	void display_connected_components_values() ;
	void display_connected_components(string message) ;
	void display_vertices_components() ;

										// predecessors, successors
	void display_predeccessors(Tvertice_identifier ident) ;
	void display_successors(Tvertice_identifier ident) ;

			  // retourne une reference vers les adjacences du noeud d'identifiant ident
	Tadjacences& get_adjacences(Tvertice_identifier ident) ;

	Tvertice_identifiers_set& get_predeccessors(Tvertice_identifier ident) ;

	Tvertice_identifiers_set& get_successors(Tvertice_identifier ident) ;


		 	 	 	 // retourne true s'il existe noeud vertice dans le graphe
	bool in_graph(Tvertice_identifier ident) ;


					 // retire le noeud vertice et tous les arcs y faisant reference
					 // (ELIMINE LES COMPOSANTES CONNEXES SI ELLES ONT ETE CALCULEES)
					 // (elles seront automatiquement recalculees si on les invoque)
	void discard_vertice(Tvertice_identifier ident) ;


		/**
		 * retourne le degre du noeud vertice (s'il existe)
		 * sinon retoune -1
		 */
	virtual float degree(Tvertice_identifier ident) ;


		 // retourne l'ensemble des noeuds connectes (adjacents) au noeud ident
	void adjacent_vertices(Tvertice_identifier ident,
							   Tvertice_identifiers_set& vertices) ;

			// retourne true si x et y sont adjacent dans le graphe courant
	bool adjacent(const Tvertice_identifier& x, const Tvertice_identifier& y) ;


		// ----------------------- DISCARD VERTICE ELEMENTS --------------------
	 	 	 // retire des elements du noeud d'identifiant identifier
	void discard_vertice_elements(Tvertice_identifier ident,
				 		  	bool adjacences, bool map, bool identifier) ;

	 	 	 // retire des elements d'un ensemble de noeuds idents
	void discard_vertices_elements(Tvertice_identifiers_set idents,
							bool adjacences, bool map, bool identifier) ;

 	 // retire des elements d'un ensemble de noeuds idents
	void discard_vertices_elements(Tvertice_identifier* idents,
						bool adjacences, bool map, bool identifier) ;


		// ----------------------- COMPOSANTES CONNEXES ------------------------

		void compute_connected_components() ;
		void discard_connected_components() ;
		void update_connected_components() ;


	 	 // retourne une reference sur les composantes connexes du graphe
		Tconnected_components& get_connected_components() ;

				// retourne les noeuds de la composante connexe du noeud ident
				// (n'ajoute aucun noeud si vertice n'est pas dans le graph)
		void connected_component(
								 Tvertice_identifier ident,
								 Tvertice_identifiers_set& remaining_vertices,
								 Tvertice_identifiers_set& component
								 ) ;

		// ---------------- (SOUS-)GRAPHE INDUIT/ REDUIT--------------------

		 	 	 	// retourne le sous-graphe induit par les noeuds idents
		void induced_subgraph(Tvertice_identifiers_set& idents, graph<Tvalue>& g) ;

		 	 	 	 // reduit le graphe courant aux noeuds et arcs impliquant les noeuds
		 	 	 	 // presents dans idents
		bool reduce_graph(Tvertice_identifiers_set& idents) ;

		 	 	 	 // efface l'ensemble des noeuds presents dans idents
		void discard_vertices(Tvertice_identifiers_set& idents) ;


		// ---------------------- DISTANCES ------------------------------------

					// retourne les sommets a distance k du sommet x
		void distance(
			   			Tvertice_identifier x,
						int k,
						Tvertice_identifiers_set& vertices,
						Tvertice_identifiers_set& dk1,		// sommets a distance k-1
						Tvertice_identifiers_set& dmaxk2	// sommets a distance au plus k-2
					 ) ;


					// retourne les sommets a distance au plus k du sommet x
		void distance_max(
			  			 	 Tvertice_identifier x,
							 int k,
							 Tvertice_identifiers_set& vertices
						 ) ;

		// ---------------------- ABSTRACTIONS ---------------------------------

		bool modularity_ok(
					  statsModularity& statsMod,
				      float modularity,
					  graph<string>* initial_graph,
					  graph<string>* c_graph,			// graph precedent dans la specialisation
					  map<Tvertice_identifier, Tvertice_identifiers_set>& c_edges,
		              bool& output_pattern) ;


		bool graph_ok(int min_sup,
					  statsModularity& statsMod,
					  bool modularity_on,
					  float modularity,
					  graph<string>* initial_graph,
					  graph<string>* c_graph,			// graph precedent dans la specialisation
					  map<Tvertice_identifier, Tvertice_identifiers_set>& c_edges,	// arcs (special mod) du graphe precedent
					  bool& output_pattern) ;

		int compute_mcb(
						graph<string>* c_graph,
						map<Tvertice_identifier, Tvertice_identifiers_set>& c_edges
					   ) ;

	 	 	 // retire les noeuds (et arcs y faisant reference) qui ne sont
	 	 	 // pas de degre au moins "degre"
		bool abstraction_minimal_degree(int min_sup, float min_degre) ;

			/*
			 * on se ramene a l'abstraction de degre minimale d'un graphe non oriente
			 * SI LES COMPOSANTES CONNEXES ONT ETE CALCULEES,
			 * ELLES SONT MISES a JOUR EN CONSEQUENCES
			 */
	    virtual bool abstraction_minimal_degree_in_out(int min_sup,
	    									            float min_in, float min_out) ;

		virtual bool abstraction_minimal_degre_simple(int min_sup,float min_degre) ;

		bool discard_isolated_vertices() ;

		void common_adjacent_vertices(
		            			 	  const Tvertice_identifier& x,
		                         	  const Tvertice_identifier& y,
								      Tvertice_identifiers_set& common_vertices
							         ) ;

		 	 	// conserve les arcs (x, y) tels que x et y ont au moins
				// k-2 voisins communs
		bool abstraction_k_dense(int min_sup, int k) ;

	 	 	 	 // conserve un noeud s'il est de degre superieur ou egale a min_degre,
				 // ou s'il est a distance au plus max_distance d'un noeud de degr
				 // superieur ou egal a min_degre
		bool abstraction_near_star(float min_degre, float max_distance) ;

		virtual bool abstraction_classic_hub_autority_pred(int min_sup, float min_hub,
														   float min_autority,
														   Tvertice_identifiers_set& H,
														   Tvertice_identifiers_set& A) ;

		virtual bool abstraction_classic_hub_autority(int min_sup, float min_hub, float min_autority) ;

		// virtual bool abstraction_hub_autority(int min_sup, int hub, int autority) ;

		virtual bool abstraction_near_star_in(float min_degre_in, float max_distance) ;

		virtual bool abstraction_near_star_out(float min_degre_out, float max_distance) ;

		 	 	 // retire les noeuds qui ne sont pas dans une composante connexe
		 	 	 // de taille au moins k
		bool abstraction_minimal_connected_components(int min_sup, int min_cc) ;

				// retire les noeuds qui ne sont pas dans un triangle
				// retourne un pointeur sur le graphe des triangles cree a cet occasion
		bool abstraction_triangles(int  min_sup) ;

			 	// abstraction topologique

		bool abstraction_topologique(int min_sup, float alpha, graph<string>* initial_graph) ;

		void apply_abstractions(abstraction& a, int min_sup, graph<string>* initial_graph, Troles_set& roles) ;


			// ------------------- CLIQUES-------------------------------------

		// retourne true s'il existe une k-clique dans vertices
	bool existe_k_clique(
			const int k,												// taille de la k-clique (k >= 1)
			Tvertice_identifiers_set& vertices,							// ensemble de noeuds ou chercher la k-clique
			Tvertice_identifiers_set& clique) ;							// clique en retour si true

} ;

/* ===================== IMPLEMENTATIONS FONCTIONS PRIVEES ================== */


// -------------------------------- GETTERS ------------------------------------

					// retourne un iterateur sur le noeud d'identifiant ident
					// dans le vertices_map
template <typename Tvalue>
Tvertices_map_iterator<Tvalue> graph<Tvalue>::
vertice_map_iterator(Tvertice_identifier ident)
{
	Tvertices_map_iterator<Tvalue> it = this->vertices_map.find(ident) ;

	if (it != this->vertices_map.end())
			return it ;

	throw string(" ERROR : ACCESS TO A NON EXISTING VERTICE!!!") ;
}


				// retourne un iterateur sur les adjacences
				// du vertice d'identifiant ident
template <typename Tvalue>
Tadjacences_map_iterator graph<Tvalue>::
adjacences_map_iterator(Tvertice_identifier ident)
{
	Tadjacences_map_iterator it = this->adjacences_map.find(ident) ;

	if (it != this->adjacences_map.end())
			return it ;

	throw string("!!! ERROR : NO ADJACENCE FOR THE CURRENT VERTICE !!!") ;
}


				// retourne une reference sur le noeud interne d'identifiant ident
				// retourne -1 si le noeud v n'est pas dans le graph
template <typename Tvalue>
Tvertice<Tvalue>& graph<Tvalue>::get_vertice(Tvertice_identifier ident)
{
	Tvertices_map_iterator<Tvalue> it = this->vertice_map_iterator(ident);
	return it->second ;
}

// ----------------------------- COMPOSANTES CONNEXES---------------------------

		// met -1 comme numero de composante connexe
		// associe a chaque noeud
template <typename Tvalue>
void graph<Tvalue>::reset_component_vertices()
{
	Tvertices_map_iterator<Tvalue> it     = this->vertices_map.begin() ;
	Tvertices_map_iterator<Tvalue> it_end = this->vertices_map.end() ;

	for ( ; it != it_end ; it++)
	{
		it->second.component = -1 ;
	}
}


// ------------------------------ ABSTRACTIONS -----------------------------


// retourne les noeuds adjacents a x et y
template <typename Tvalue>
void graph<Tvalue>::
common_adjacent_vertices(
            			 const Tvertice_identifier& x,
                         const Tvertice_identifier& y,
						 Tvertice_identifiers_set& common_vertices
					    )
{
	Tvertice_identifiers_set adjacent_x ;
	Tvertice_identifiers_set adjacent_y ;


	this->adjacent_vertices(x, adjacent_x) ;
	this->adjacent_vertices(y, adjacent_y) ;

	// displaySet("adja_x", adjacent_x) ;
	// displaySet("adja_y", adjacent_y) ;

	// setIntersection(&common_vertices, adjacent_x, adjacent_y);

	set_intersection (
			             adjacent_x.begin(), adjacent_x.end(),
						 adjacent_y.begin(), adjacent_y.end(),
			             inserter(common_vertices,common_vertices.begin())
					    ) ;

	// displaySet("common_vertices", common_vertices) ;
}


						// retourne true si les noeuds x et y ont au moins k
						// voisins communs
template <typename Tvalue>
bool graph<Tvalue>::sufficient_common_adjacent_vertices(
											            const Tvertice_identifier& x,
		                                                const Tvertice_identifier& y,
											            int k
											           )
{
	Tvertice_identifiers_set common_vertices ;
	this->common_adjacent_vertices(x, y, common_vertices) ;

	return common_vertices.size() >= k ;
}


/* ================== IMPLEMENTATIONS FONCTIONS PUBLIQUES =================== */


// -------------------------------- CONSTRUCTEURS ------------------------------


// construit le graphe avec un nom et une fonction d'affichage
template <typename Tvalue>
graph<Tvalue>::graph()
{
}

			// construit un graphe vide avec un nom et une fonction d'affichage
template <typename Tvalue>
graph<Tvalue>::graph(
		             string aName,
				     function<void (Tvalue)> display_vertice_value_function,
					 function<string (Tvalue)> toString_value_function
					)
{
	this->name 		    = aName ;
	this->display_value = display_vertice_value_function ;
	this->toString_value = toString_value_function ;
}


/**
 * construit le graphe a partir d'une liste de noeuds et
 * d'adjacences (avec predecesseurs et successeurs).
 * ATTENTION : les objets internes sont des copies des parama�tres
 */
template <typename Tvalue>
graph<Tvalue>::
graph(
		string aName,
		Tvertices_map<Tvalue>& vertices,
		Tadjacences_map& adjacences,
		function<void (Tvalue)> display_vertice_value_function,
		function<string (Tvalue)> toString_value_function
	  )
{
	this->name 					      = aName ;
	this->vertices_map 			      = vertices ;			// copie
	this->adjacences_map 		      = adjacences ;		// copie
	this->display_value 			  = display_vertice_value_function ;
	this->toString_value 			  = toString_value_function ;
	this->build_vertice_identifiers() ;
}

	// construit un graph avec des noeuds et sans adjacences
template <typename Tvalue>
graph<Tvalue>::
graph(
		string aName,
		Tvertices_map<Tvalue>& vertices,
		function<void (Tvalue)> display_vertice_value_function,
		function<string (Tvalue)> toString_value_function
	  )
{
	this->name 					      = aName ;
	this->vertices_map 			      = vertices ;			// copie
	this->display_value				  = display_vertice_value_function ;
	this->toString_value 			  = toString_value_function ;
	this->build_vertice_identifiers() ;
}


template <typename Tvalue>
void graph<Tvalue>::build_vertice_identifiers()
{
	Tvertices_map_iterator<Tvalue> it     = this->vertices_map.begin() ;
	Tvertices_map_iterator<Tvalue> it_end = this->vertices_map.end()   ;

	for ( ; it != it_end ; it++)
	{
		this->vertice_identifiers_set.insert(it->first) ;
	}
}


	// DESTRUCTEUR 

template <typename Tvalue>
graph<Tvalue>::~graph()
{}


// --------------------------- COPIE DU GRAPHE ----------------------------

// constructeur par copie
template <typename Tvalue>
graph<Tvalue>::graph(const graph<Tvalue>& g)
{
	*this = g ;
}


template <typename Tvalue>
graph<Tvalue>* graph<Tvalue>::copy_graph()
{
	return new graph<Tvalue>(*this) ;
}

// ------------------------- SAUVEGARDE -----------------------------------

// sauvegarde le graphe en format dot
template <typename Tvalue>
void graph<Tvalue>::visu_graph_dot(string graph_name,
								   bool vertice_ident,
								   bool vertice_label)
{
	this->save_graph_dot(graph_name, vertice_ident, vertice_label) ;
	string command =  "xdot ./" + graph_name + ".dot &" ;
	// string prefix = ""
	// string command =  prefix + "dotty ./" + graph_name + ".dot" ;
	// cout << "command = " << command << endl ;

	int x = std::system(command.c_str()) ;
}


// sauvegarde le graph en format dot
template <typename Tvalue>
void graph<Tvalue>::save_graph_dot(string name, bool vertice_ident, bool vertice_label)
{
	this->save_graph_dot2(name, vertice_ident, vertice_label, false) ;
}


// sauvegarde le graphe en format dot
template <typename Tvalue>
void graph<Tvalue>::save_graph_dot2(string graph_name,
									bool vertice_ident,
                                    bool vertice_label,
									bool directed_graph)
{
	string file_name = "./" + graph_name + ".dot" ;
	ofstream fichier(file_name, ios::out | ios::trunc);

	string kind_of_graph = "Graph";
	string kind_of_edge  = " -- ";

	if (directed_graph)
	{
		kind_of_graph = "diGraph" ;
		kind_of_edge  = " -> " ;
	}

	if(fichier)
	{
		fichier << kind_of_graph  + " " + graph_name << "{" << endl ;
					 // sauve les noeuds (vertices)

											// titre du graph
		fichier << "labelloc=top;" << endl ;
		// fichier << "labeljust=left;" << endl ;
		fichier << "fontsize=20;" << endl ;
		fichier << "label=" << "\"" << graph_name << "\";" << endl ;

					 // pour chaque noeud x
	for (Tvertice_identifiers_set_iterator x = this->vertice_identifiers_set.begin() ;
			x != this->vertice_identifiers_set.end() ; x++)
	{
		string nom_noeud = graph<Tvalue>::toString_value(this->get_vertice_value(*x)) ;
		string identifiant_noeud = to_string(*x) ;
		string label_noeud = get_vertice_label(*x) ;

		string label_node = nom_noeud ;

		if (vertice_ident)
		{
			label_node = identifiant_noeud + ":" + label_node ;
		}

		if (vertice_label)
		{
			label_node = label_node + ":" + label_noeud ;
		}


		string option_label = identifiant_noeud + " [label= \"" + label_node + "\"" + "]" ;

		fichier << option_label << ";" << endl ;
	}

					// sauve les aretes (ici arcs)

						// pour chaque noeud x
	for (Tvertice_identifiers_set_iterator x = this->vertice_identifiers_set.begin() ;
			x != this->vertice_identifiers_set.end() ; x++)
	{
		Tvertice_identifiers_set& successors_x = this->get_successors(*x) ;
		Tvertice_identifiers_set_iterator y    = successors_x.begin() ;

							// pour chaque successeur y de x
		for ( ; y != successors_x.end() ; y++)
		{
			if (vertice_ident)
				fichier << to_string(*x) << kind_of_edge << to_string(*y) << " ;" << endl ;
			else
				fichier << graph<Tvalue>::toString_value(this->get_vertice_value(*x))
						<< graph<Tvalue>::toString_value(this->get_vertice_value(*y))
						<< " ;" << endl ;
		}
	}

	fichier << "}" << endl ;
	fichier.close();
}
else
cerr << "Erreur a l'ouverture !" << endl;
}




// ------------------------------ GETTERS --------------------------------------

template <typename Tvalue>
Tvertice_identifiers_set& graph<Tvalue>::get_vertice_identifiers_list()
{
	return this->vertice_identifiers_set ;
}


template <typename Tvalue>
Tvertice_identifiers_set graph<Tvalue>::get_copy_vertice_identifiers_list()
{
	return this->vertice_identifiers_set ;
}


template <typename Tvalue>
string graph<Tvalue>:: get_graph_name()
{
	return this->name ;
}

		// retourne une reference sur la valeur du noeud interne d'identifiant ident
template <typename Tvalue>
Tvalue& graph<Tvalue>::get_vertice_value(Tvertice_identifier ident)
{
    return this->vertice_map_iterator(ident)->second.value ;
}


		// retourne une reference sur liste d'adjacences constituant le graphe
template <typename Tvalue>
Tvalue& graph<Tvalue>::get_vertice_value(Tvertices_map_iterator<Tvalue> it)
{
	return it->second.value ;
}


	// retourne une reference sur le label
	// associe au noeud d'identifiant ident
template <typename Tvalue>
string& graph<Tvalue>::get_vertice_label(Tvertice_identifier ident)
{
	Tvertices_map_iterator<Tvalue> it = this->vertice_map_iterator(ident) ;
	return it->second.label ;
}

	// retourne une reference sur le label
	// associe au noeud pointe par it
template <typename Tvalue>
string& graph<Tvalue>::get_vertice_label(Tvertices_map_iterator<Tvalue> it)
{
	return it->second.label ;
}

 	 	 	 	 // retourne une reference sur la liste de noeuds du graphe
template <typename Tvalue>
Tvertices_map<Tvalue>& graph<Tvalue>::get_vertices_map()
{
	return this->vertices_map ;
}


// retourne une reference sur liste d'adjacences constituant le graphe
template <typename Tvalue>
Tadjacences_map& graph<Tvalue>::get_adjacences_map()
{
	return this->adjacences_map ;
}


 	  // retourne une reference vers les adjacences du noeud d'identifiant ident
template <typename Tvalue>
Tadjacences& graph<Tvalue>::get_adjacences(Tvertice_identifier ident)
{
	Tadjacences_map_iterator it = this->adjacences_map_iterator(ident) ;
	return it->second ;
}


	// retourne la liste des predecesseurs du noeud vertice
template <typename Tvalue>
Tvertice_identifiers_set& graph<Tvalue>::get_predeccessors(Tvertice_identifier ident)
{
	Tadjacences_map_iterator it = this->adjacences_map_iterator(ident) ;
	return it->second.predeccessors ;
}


 	 	 	 	 	 	 // retourne la liste des successeurs du noeud vertice
template <typename Tvalue>
Tvertice_identifiers_set& graph<Tvalue>::get_successors(Tvertice_identifier ident)
{
	Tadjacences_map_iterator it = this->adjacences_map_iterator(ident) ;
	return it->second.successors ;
}


// ---------------------------- SETTERS ------------------------------------

template <typename Tvalue>
void graph<Tvalue>::set_adjacences(Tadjacences_map& an_adjacences_map)
{
	this->adjacences_map = an_adjacences_map ;	// attention copie
}


template <typename Tvalue>
void graph<Tvalue>::set_vertices(Tvertices_map<Tvalue>& a_vertices_map)
{
	this->vertices_map = a_vertices_map ;				// attention copie
	this->build_vertice_identifiers() ;
}

/* ------------------------- INFORMATIONS SUR LE GRAPHE --------------------- */

template <typename Tvalue>
int graph<Tvalue>::nb_vertices()
{
	return this->vertice_identifiers_set.size() ;
}


		// retourne un map dont chaque element <ident, degre>
		// contient le degre du noeud d'identifiant ident
template <typename Tvalue>
void graph<Tvalue>::compute_map_degres()
{
								//conserver les degres de chaque sommet
									// dans degres
	Tvertice_identifiers_set vertices = this->get_vertice_identifiers_list() ;		

	Tvertice_identifiers_set_iterator it     = vertices.begin() ;
	Tvertice_identifiers_set_iterator it_end = vertices.end() ;

	for ( ; it != it_end ; it++)
	{
		this->map_degres[*it] = this->degree(*it) ;
			
	}							
}


		// retourne le degre initial du noeud d'identifiant ident
template <typename Tvalue>
float graph<Tvalue>::get_map_degre_vertice(Tvertice_identifier ident)
{
	return this->map_degres[ident] ;
}

		// retourne la map des degres initiaux des noeuds du graphe
template <typename Tvalue>
map<Tvertice_identifier, float>&  graph<Tvalue>::get_map_degres()
{
	return this->map_degres ;
}


/* retourne le nombre d'arc :
 * (ATTENTION : en interne un graphe non oriente contient l'arc (a-b) et l'arc (b-a)
 *  ici on ne compte qu'un arc en ne parcourant que les successeurs.
 *  Le comptage est donc correct pour les graphes orientes ou non
 */

template <typename Tvalue>

int graph<Tvalue>::nb_edges()
{
	int nb_edges = 0 ;

	Tadjacences_map_iterator it     = this->adjacences_map.begin() ;
	Tadjacences_map_iterator it_end = this->adjacences_map.end() ;

	for ( ; it != it_end ; it++)
	{
		nb_edges += (*it).second.successors.size() ;
	}

	return nb_edges ;
}


/* ----------------------------- AFFICHAGE ---------------------------------- */

				// vertice

template <typename Tvalue>
void graph<Tvalue>::display_vertice_value(Tvertice_identifier ident)
{
	Tvalue v = this->get_vertice_value(ident) ;
	this->display_value(v) ;
}


			// affiche le noeud pointe par it
template <typename Tvalue>
void graph<Tvalue>::display_vertice_value_it(Tvertices_map_iterator<Tvalue> it)
{
	Tvalue v = this->get_vertice_value(it) ;
	this->display_value(v) ;
}


		// affiche le noeud interne pointe par it
template <typename Tvalue>
void graph<Tvalue>::display_vertice_it(Tvertices_map_iterator<Tvalue> it)
{
	cout << "identifier          : " << it->first << endl ;
	cout << "value               : " ;
	this->display_value(it->second.value) ;
	cout << endl ;
	cout << "label               : " << it->second.label << endl ;
	cout << "connected component : " << it->second.component << endl ;
}


			// affiche le noeud interne d'identifiant ident
template <typename Tvalue>
void graph<Tvalue>::display_vertice(Tvertice_identifier ident)
{

	Tvertice<Tvalue>& v = this->get_vertice(ident) ;

	cout << "internal vertice       : " << ident << endl ;
	cout << "value               : " ;
	this->display_value(v.value) ;
	cout << endl ;
	cout << "label               : " << v.label << endl ;
	cout << "connected component : " << v.component << endl ;
}


 	 	 	 	 // affiche les predecesseurs d'un noeud interne d'identificateur ident
template <typename Tvalue>
void graph<Tvalue>::display_predeccessors(Tvertice_identifier ident)
{
	graph::display_vertices_set(this->get_predeccessors(ident)) ;
}


/**
 * affiche les successeurs d'un noeud
 */
template <typename Tvalue>
void graph<Tvalue>::display_successors(Tvertice_identifier ident)
{
	graph::display_vertices_set(this->get_successors(ident)) ;
}


/**
 *  display the graph
 */
template <typename Tvalue>
void graph<Tvalue>::display_graph(string message)
{
	cout << message << endl << endl ;
	this->display_graph() ;
}


/**
 *  display the graph
 */
template <typename Tvalue>
void graph<Tvalue>::display_graph()
{
	cout << "graph    : " << this->get_graph_name() << endl ;

	this->display_vertices() ;
	this->display_edges() ;
}


// display the set of set of vertices identifies par ident_set
template <typename Tvalue>
void graph<Tvalue>::display_vertices_set()
{
	graph<Tvalue>::display_vertices_set(this->get_vertice_identifiers_list()) ;
}


// display the set of set of vertices identifies par ident_set
template <typename Tvalue>
void graph<Tvalue>::display_vertices_set(const Tvertice_identifiers_set& idents_set)
{
	cout << "{" ;

	Tvertice_identifiers_set_iterator it     = idents_set.begin() ;
	Tvertice_identifiers_set_iterator it_end = idents_set.end() ;

	if (it!= it_end)
	{
		// cout << "*it = " << *it << endl ;
		this->display_vertice_value(*it) ;
		it++ ;
	}

	while(it != it_end)
	{
		cout << ", " ;
		// cout << "*it = " << *it << endl ;
		this->display_vertice_value(*it) ;
		it++ ;
	}

	cout << "}" ;
}


	// display each vertice of the set with
template <typename Tvalue>
void graph<Tvalue>::
display_vertices_idents_values_set(Tvertice_identifiers_set& vertices)
{
	cout << "{" ;

		Tvertice_identifiers_set_iterator it     = vertices.begin() ;
		Tvertice_identifiers_set_iterator it_end = vertices.end() ;

		if (it!= it_end)
		{
			cout << "<" << *it << " ; " ;
			this->display_vertice_value(*it) ;
			cout << ">" ;
			it++ ;
		}

		while(it != it_end)
		{
			cout << ", <" << *it << " ; " ;
			this->display_vertice_value(*it) ;
			cout << ">" ;
			it++ ;
		}

		cout << "}" ;
}



	// retourne un string contenant une liste d'identifiants de noeuds
template <typename Tvalue>
string graph<Tvalue>::toString_vertice_identifiers_set(Tvertice_identifiers_set idents_set)
{
	string s = "" ;

	s = s + "{" ;

		Tvertice_identifiers_set_iterator it     = idents_set.begin() ;
		Tvertice_identifiers_set_iterator it_end = idents_set.end() ;

		if (it!= it_end)
		{
			s = s +  to_string(*it) ;
			it++ ;
		}

		while(it != it_end)
		{
			s = s + ", " ;
			s = s + to_string(*it) ;
			it++ ;
		}

		s = s + "}" ;

		return s ;
}


// display the set of set of vertices identifies par ident_set
template <typename Tvalue>
void graph<Tvalue>::display_vertice_identifiers_set(Tvertice_identifiers_set idents_set)
{
	cout << "{" ;

	Tvertice_identifiers_set_iterator it     = idents_set.begin() ;
	Tvertice_identifiers_set_iterator it_end = idents_set.end() ;

	if (it!= it_end)
	{
		cout << *it ;
		it++ ;
	}

	while(it != it_end)
	{
		cout << ", " ;
		cout << *it ;
		it++ ;
	}

	cout << "}" ;
}


					// display all infos of each vertice
template <typename Tvalue>
void graph<Tvalue>::display_vertices()
{

	cout << "vertices : " << endl ;

	Tvertices_map_iterator<Tvalue> it     = this->vertices_map.begin() ;
	Tvertices_map_iterator<Tvalue> it_end = this->vertices_map.end() ;

	if (it!= it_end)
	{
		this->display_vertice_it(it) ;

		it++ ;

		while(it != it_end)
		{
			cout <<  endl ;

			this->display_vertice_it(it) ;

			it++ ;
		}
	}

	cout << endl  ;
}


// display all infos of each vertice
template <typename Tvalue>
void graph<Tvalue>::display_vertice_identifiers()
{
	cout << "vertices : {"  ;

	Tvertice_identifiers_set_iterator it 	 = this->vertice_identifiers_set.begin() ;
	Tvertice_identifiers_set_iterator it_end = this->vertice_identifiers_set.end() ;

	if (it != it_end)
	{
		cout << *it ;
		it++ ;
	}

	while(it != it_end)
	{
		cout << ", " << *it ;

		it++ ;
	}

	cout << "}" << endl  ;
}


template <typename Tvalue>
void graph<Tvalue>::display_vertices_values()
{
	cout << "vertices : " ;
	cout << "{" ;

	Tvertices_map_iterator<Tvalue> it     = this->vertices_map.begin() ;
	Tvertices_map_iterator<Tvalue> it_end = this->vertices_map.end() ;

	if (it!= it_end)
	{
		this->display_vertice_value(it) ;

		it++ ;

		while(it != it_end)
		{
			cout << ", " ;

			this->display_vertice_value(it) ;

			it++ ;
		}
	}

	cout << "}" << endl ;
}


template <typename Tvalue>
void graph<Tvalue>::display_edges()
{
	 cout << "edges    :" << endl ;
	 Tadjacences_map_iterator it     = this->adjacences_map.begin() ;
	 Tadjacences_map_iterator it_end = this->adjacences_map.end() ;

	 	 	 	 	 	 	 	 	 // pour chaque noeud du graphe
	while(it != it_end)
	{
		Tvalue source = this->get_vertice_value(it->first) ;

		cout << "vertice <" << it->first << " ; ";
		this->display_value(source) ;
		cout << ">" << endl ;

		cout << "    predeccessors : " ;
		graph<Tvalue>::display_vertice_identifiers_set(it->second.predeccessors) ;
		cout << endl ;

		cout << "    successors    : " ;
		graph<Tvalue>::display_vertice_identifiers_set(it->second.successors) ;
		cout << endl ;

		it++ ;
	}
}


// ----------------------------- GESTION NOEUD ---------------------------------

// retourne true si x et y sont adjacent dans le graphe courant
template <typename Tvalue>
bool graph<Tvalue>::adjacent(const Tvertice_identifier& x,
							 const Tvertice_identifier& y)
{
					// verifie si y fait partie des precedesseurs ou
					// successeurs de x
	return this->get_predeccessors(x).find(y) != this->get_predeccessors(x).end() ||
		   this->get_successors(x).find(y)    != this->get_successors(x).end() ;
}



	// retourne l'ensemble des noeuds connectes au noeud ident
template <typename Tvalue>
void graph<Tvalue>::
adjacent_vertices(Tvertice_identifier ident, Tvertice_identifiers_set& vertices)
{
	vertices.insert(this->get_predeccessors(ident).begin(),
					this->get_predeccessors(ident).end()) ;

	vertices.insert(this->get_successors(ident).begin(),
					this->get_successors(ident).end()) ;
}


/**
 * retire le noeud vertice et tous les arcs
 * NE MODIFIE PAS LES COMPOSANTES CONNEXES CONTENANT CE NOEUD
 * (POUR LE MOMENT)
 */
template <typename Tvalue>
void graph<Tvalue>::discard_vertice(Tvertice_identifier ident)
{
	this->discard_vertice_elements(ident, true, true, true) ;
}


/**
 * retire des elements du noeud d'identifiant ident
 * NE MODIFIE PAS LES COMPOSANTES CONNEXES CONTENANT CE NOEUD
 * (POUR LE MOMENT)
 */
template <typename Tvalue>
void graph<Tvalue>::discard_vertice_elements(
											   Tvertice_identifier ident,
											   bool adjacences,
											   bool map,
											   bool identifier
											  )
{
	// cout << "ident = " << ident << endl ;

	if (adjacences)
			this->discard_vertice_adjacences(ident) ;

	if (map)
			this->vertices_map.erase(ident) ;

	if (identifier)
		this->vertice_identifiers_set.erase(ident) ;

	if (this->connected_components_computed)
		this->recompute_connected_components = true ;
}

					// elimine toutes les adjacences faisant reference a� ident
template <typename Tvalue>
void graph<Tvalue>::discard_vertice_adjacences(Tvertice_identifier ident)
{
	if (!this->in_graph(ident))
	{
			cerr << "node " << ident << " not in graph" << endl ;
			return ;
	}

		Tadjacences&  adja = this->get_adjacences(ident) ;

										// on efface vertice
										// dans les predecesseurs de chaque
										// noeud successeur de vertice ;

		Tvertice_identifiers_set& successors   = adja.successors    ;
		Tvertice_identifiers_set_iterator it_s = successors.begin() ;

										// pour chaque idenfidiant de noeud *it_s successeur de ident
		while (it_s != successors.end())
		{
			    // cout << "succ = *it_s = " << *it_s << endl ;

										// on efface ident des predecesseurs de *it_s
				this->get_adjacences(*it_s).predeccessors.erase(ident) ;
				it_s++ ;
		}

		// on efface vertice
		// dans les successeurs de chaque noeud predecesseur de vertice ;

		Tvertice_identifiers_set& predecessors = adja.predeccessors    ;
		Tvertice_identifiers_set_iterator it_p = predecessors.begin() ;

										// pour chaque noeud *it_p predecesseur de vertice
		while (it_p != predecessors.end())
		{
				// cout << "pred = *it_p = " << *it_p << endl ;
										// on efface vertice de la liste des successeurs de *it
				this->get_adjacences(*it_p).successors.erase(ident) ;
				it_p++ ;
		}
										// on efface la liste des successeurs
										// et predecesseurs de ident
		this->adjacences_map.erase(ident)  ;
}


 	 	 // retourne true s'il existe un noeud d'identifiant ident dans le graph
template <typename Tvalue>
bool graph<Tvalue>::in_graph(Tvertice_identifier ident)
{
	if (this->vertice_identifiers_set.find(ident) == this->vertice_identifiers_set.end())
	{
		cout << "no vertice with identier " << ident << " in the graph" << endl ;
		return false ;
	}
	return true ;
}


/**
 * retourne le degre du noeud vertice (s'il existe)
 * sinon retoune -1
 */

template <typename Tvalue>
float graph<Tvalue>::degree(Tvertice_identifier ident)
{
	if (!this->in_graph(ident))
		return -1 ;

	return this->get_successors(ident).size() + this->get_predeccessors(ident).size() ;
}

// ------------------------------- COMPOSANTES CONNEXES ------------------------

			// retourne une reference sur le numero de composante
			// associe au noeud d'identifiant ident
template <typename Tvalue>
int& graph<Tvalue>::get_vertice_component(Tvertice_identifier ident)
{
	Tvertices_map_iterator<Tvalue> it = this->vertice_map_iterator(ident) ;
	return it->second.component ;
}

			// retourne une reference sur le numero de composante
			// associe au noeud pointe par it
template <typename Tvalue>
int& graph<Tvalue>::get_vertice_component(Tvertices_map_iterator<Tvalue> it)
{
	return it->second.component ;
}
 	 	 // retourne une reference sur les composantes connexes du graphe

template <typename Tvalue>
Tconnected_components& graph<Tvalue>::get_connected_components()
{
	return this->connected_components ;
}

template <typename Tvalue>
void graph<Tvalue>::display_connected_components(string message)
{
	cout << message << endl ;
	this->display_connected_components() ;
}


template <typename Tvalue>
void graph<Tvalue>::display_connected_components()
{
	if (!this->connected_components_computed)
		this->compute_connected_components() ;

	cout << "connected components of the graph " << this->get_graph_name() << " :" << endl ;
	cout << "{" << endl;

	Tconnected_components_iterator it     = this->get_connected_components().begin() ;
	Tconnected_components_iterator it_end = this->get_connected_components().end() ;

	if (it!= it_end)
	{
		cout << "  " << it->first << " : " ;				// numero composante
		display_vertices_set(it->second) ;
		it++ ;
	}

	while(it != it_end)
	{
		cout << ", "  << endl << "  " << it->first << " : " ;
		display_vertices_set(it->second) ;
		it++ ;
	}

	cout << endl << "}" << endl ;
}


template <typename Tvalue>
void graph<Tvalue>::display_connected_components_values()
{
	if (!this->connected_components_computed)
		this->compute_connected_components() ;

	cout << "connected components of the graph " << this->get_graph_name() << " :" << endl ;
	cout << "{" << endl;

	Tconnected_components_iterator it     = this->get_connected_components().begin() ;
	Tconnected_components_iterator it_end = this->get_connected_components().end() ;

	if (it!= it_end)
	{
		cout << "  " << it->first << " : " ;				// numero composante
		display_vertices_idents_values_set(it->second) ;
		it++ ;
	}

	while(it != it_end)
	{
		cout << ", "  << endl << "  " << it->first << " : " ;
		display_vertices_idents_values_set(it->second) ;
		it++ ;
	}

	cout << endl << "}" << endl ;
}


/**
 * display the pairs (vertice - component)
 */
template <typename Tvalue>
void graph<Tvalue>::display_vertices_components()
{
		if (!this->connected_components_computed)
			this->compute_connected_components() ;

		cout << "{" << endl;

		Tvertices_map_iterator<Tvalue> it     = this->vertices_map.begin() ;
		Tvertices_map_iterator<Tvalue> it_end = this->vertices_map.end() ;
		Tvertice_identifier ident ;

		if (it!= it_end)
		{
			ident = it->first ;

			cout << "  (" << ident << ", " << this->get_vertice_component(ident) << ")" ;
			it++ ;
		}

		while(it != it_end)
		{
			ident = it->first ;

			cout << ", "  << endl << "  (" << ident << ", " << this->get_vertice_component(ident) << ")" ;
			it++ ;
		}

		cout << endl << "}" << endl ;
}


/**
* calcule les composantes connexes du graphe
*/
template <typename Tvalue>
void graph<Tvalue>::compute_connected_components()
{
	if (this->recompute_connected_components)
		this->update_connected_components() ;

	if (!this->connected_components_computed)
	{
		Tvertice_identifiers_set remaining_vertices = this->vertice_identifiers_set ; // copie
		internal_connected_components(remaining_vertices) ;
		this->connected_components_computed = true ;
		this->recompute_connected_components = false ;
	}
}


/**
* retourne les composantes connexes du graph
*/
template <typename Tvalue>
void graph<Tvalue>::
internal_connected_components(Tvertice_identifiers_set& remaining_vertices)
{
	Tvertice_identifiers_set_iterator it = remaining_vertices.begin() ;

	while (it != remaining_vertices.end())
	{
		// cerr << "composante connexe en cours : " << this->next_connected_component_ident << endl ;

									// creation composante connexe du noeud *it
		Tvertice_identifiers_set component ;

		// cout << "avant composante " << endl ;
		// this->display_graph() ;

									// calcule la composante connexe du noeud *it
		this->connected_component(*it, remaining_vertices, component) ;

		// cout << "apres composante " << endl ;
		// this->display_graph() ;


		// graph<Tvalue>::display_vertice_identifiers_set(component) ;
		// cout << endl ;

									// creation nouvelle composante connexe
		this->connected_components[this->next_connected_component_ident] = component ;

									// mis a jour prochain numero de composante connexe
		this->next_connected_component_ident++ ;

									// le noeud *it ayant ete elimine de remaining_vertices
									// on passe au noeud suivante (premier noeud restant)
		it = remaining_vertices.begin() ;
	}
}



/**
 * retourne les identifiants de la composante connexe du noeud ident
 * (n'ajoute aucun noeud si ident n'est pas un sommet du graphe)
*/
template <typename Tvalue>
void graph<Tvalue>::
connected_component(
					 Tvertice_identifier ident,
					 Tvertice_identifiers_set& remaining_vertices,
					 Tvertice_identifiers_set& component
					)
{
	Tvertice_identifier node ;
	stack<Tvertice_identifier> s ;

	s.push(ident) ;						// ident noeud de depart

	do
	{
		node = s.top() ;				// lit dessus de pile
		s.pop() ;						// elimine dessus de pile

		if (remaining_vertices.find(node) != remaining_vertices.end())
		{
										// node trouve dans les noeuds a� traiter
			// cout <<  "node : " << node << endl ;

			component.insert(node) ;

			this->get_vertice(node).component = this->next_connected_component_ident ;
			remaining_vertices.erase(node) ;

											// insere les successeurs de node
			Tvertice_identifiers_set& successors          = this->get_successors(node) ;
			Tvertice_identifiers_set_iterator it_succ     = successors.begin() ;
			Tvertice_identifiers_set_iterator it_succ_end = successors.end() ;

			for ( ; it_succ != it_succ_end ; it_succ++)
			{
				// cout << "*it = " << *it_succ << endl ;
				s.push(*it_succ) ;
			}

			// cout << "succ" << endl ;
			// graph<Tvalue>::display_vertice_identifiers_set(successors) ;
			// cout << endl ;

											// insere les successeurs de node
			Tvertice_identifiers_set& predecessors        = this->get_predeccessors(node) ;
			Tvertice_identifiers_set_iterator it_pred     = predecessors.begin() ;
			Tvertice_identifiers_set_iterator it_pred_end = predecessors.end() ;

			for ( ; it_pred != it_pred_end ; it_pred++)
			{
				// cout << "*it = " << *it_pred << endl ;
				s.push(*it_pred) ;
			}

			// cout << "pred "<< endl ;
			// graph<Tvalue>::display_vertice_identifiers_set(predecessors) ;
			// cout << endl ;
		}
	}
	while(!s.empty()) ;
}

template <typename Tvalue>
void graph<Tvalue>::discard_connected_components()
{
							// reinitilisation des variables associees aux
							// composantes connexes
		this->connected_components.clear() ;
		this->next_connected_component_ident = 0 ;
		this->reset_component_vertices() ;
		this->connected_components_computed  = false ;
		this->recompute_connected_components = false ;
}


/**
* force le calcul des composantes connexes
*/
template <typename Tvalue>
void graph<Tvalue>::update_connected_components()
{
	this->discard_connected_components() ;
	this->compute_connected_components() ;
}


/* ---------------------------- SOUS-GRAPHE INDUIT -------------------------- */

/**
 * retourne g le sous-graphe induit par les noeuds identifies par les sommets idents
 */
template <typename Tvalue>
void graph<Tvalue>::
induced_subgraph(Tvertice_identifiers_set& idents, graph<Tvalue>& g)
{
	g = *this ;
							// reduction du graph g
	g.reduce_graph(idents) ;
}


/**
 * reduit le graphe courant aux noeuds et arcs impliquant les noeuds
 * presents dans idents
 */
template <typename Tvalue>
bool graph<Tvalue>::
reduce_graph(Tvertice_identifiers_set& idents)
{
	  // graph::display_vertice_identifiers_set(idents) ;
	  // cout << endl ;
	  // graph::display_vertice_identifiers_set(this->vertice_identifiers_set) ;
	  // cout << endl ;


	Tvertice_identifiers_set diff ;

	// cerr << "avant set " << endl ;
	set_difference(
					  this->vertice_identifiers_set.begin(), this->vertice_identifiers_set.end(),
					  idents.begin(), idents.end(),
					  inserter(diff, diff.begin())
				   ) ;

	  // graph::display_vertice_identifiers_set(diff) ;
	  // cout << endl ;

	// cerr << "apres set " << endl ;

	  	  	  	  	  	  	  	  	 // diff contient les noeuds a� retirer
	if (diff.size() > 0)
	{
	  this->discard_vertices(diff) ;
	  return true ;
	}

	return false ;
}
// end reduce_graph


/**
 *   efface l'ensemble des noeuds presents dans vertices
 *   TOUTES LES STRUCTURES INTERNES DONT LES EVENTUELLES COMPOSANTES CONNEXES
 *   SONT MODIFIEES EN CONSEQUENCE
 */
template <typename Tvalue>
void graph<Tvalue>::discard_vertices(Tvertice_identifiers_set& idents)
{
	  	  	  	 // vertices contient les noeuds a� retirer

	this->discard_vertices_elements(idents, true, true, true) ;
}


/**
 *   efface l'ensemble des noeuds presents dans vertices
 *   TOUTES LES STRUCTURES INTERNES DONT LES EVENTUELLES COMPOSANTES CONNEXES
 *   SONT MODIFIEES EN CONSEQUENCE
 */
template <typename Tvalue>
void graph<Tvalue>::discard_vertices_elements(Tvertice_identifiers_set idents,
											  bool adjacences, bool map,
											  bool identifier)
{
	  	  	  	 // vertices contient les noeuds a� retirer

	Tvertice_identifiers_set_iterator it ;

	for(it = idents.begin() ; it != idents.end() ; it++)
	{
		// cout << "discard : " << *it << endl ;
		this->discard_vertice_elements(*it, adjacences, map, identifier) ;
	}
}


// ------------------------------ DISTANCES ------------------------------------

		// retourne D(x,k) l'ensemble des sommets se trouvant a� distance k
		// du sommet x
		// on utilise les proprietes suivantes :
		// D(x, 0) = {x}
		// D(x, 1) = {y | y directement connecte a� x}
		// et pour k >=2 :
		// D(x, k) = {y | il existe z dans D(x,k-1) tel que
        //                y connecte a� z et y n'est pas dans Dmax(x, k-2)

template <typename Tvalue>
void graph<Tvalue>::distance(
							Tvertice_identifier x,
							int k,
							Tvertice_identifiers_set& vertices,
							Tvertice_identifiers_set& dk1,		// noeuds a� distance k-1
							Tvertice_identifiers_set& dmaxk2	// noeuds a� distance au plus k-2
						   )
{
	if (k==0)
	{
		vertices.insert(x) ;
		return ;
	}

	if (k==1)
	{
		this->adjacent_vertices(x, vertices) ;
		return ;
	}
									// k >= 2

	Tvertice_identifiers_set_iterator it_z     = dk1.begin() ;
	Tvertice_identifiers_set_iterator it_z_end = dk1.end() ;

									// pour chaque noeud z a� distance k-1 de x
	for ( ; it_z != it_z_end ; it_z++)
	{
									// chaque noeud y connecte a z (pas deja traite)
									// qui n'est pas a une distance inferieure a� k de x
									// est a distance k de x

									// on explore les successeurs de z
		Tvertice_identifiers_set_iterator it_y     = this->get_successors(*it_z).begin() ;
		Tvertice_identifiers_set_iterator it_y_end = this->get_successors(*it_z).end() ;

		for ( ; it_y != it_y_end ; it_y++)
		{
									// on ajoute le noeud y s'il n'est pas
									// dans dmaxk2
									// s'il est deja� dans vertices
									// l'insertion ne se fera pas (set)
			if (dmaxk2.find(*it_y) == dmaxk2.end())
			{
				vertices.insert(*it_y) ;
			}
		}
									// on explore les predecesseurs de z
		it_y     = this->get_predeccessors(*it_z).begin() ;
		it_y_end = this->get_predeccessors(*it_z).end() ;

		for ( ; it_y != it_y_end ; it_y++)
		{
							// on ajoute le noeud y s'il n'est pas deja
							// present dans vertices et s'il n'est pas
							// dans dk-1
			if (vertices.find(*it_y) == vertices.end() &&
				dmaxk2.find(*it_y) == dmaxk2.end())
			{
				vertices.insert(*it_y) ;
			}
		}
	}
}

		// retourne l'ensemble des sommets se trouvant a� distance
		// au plus n du sommet x
		// on utilise les proprietes suivantes :
		// Dmax(x, 0) = {x}
		// pour k >= 1
		// Dmax(x, k) = Dmax(x, k-1) union D(x, k)
		// (D(x, k) defini dans le fonction precedente)
		// on utilise ici une fonction iterative (non recursive)

template <typename Tvalue>
void graph<Tvalue>::distance_max(
								 Tvertice_identifier x,
								 int n,
								 Tvertice_identifiers_set& vertices
								)
{
		vertices.insert(x) ;					// vertices contient le noeud
												// a distance 0 de x
		if (n == 0)
			return ;

		Tvertice_identifiers_set* dk1 = new Tvertice_identifiers_set() ;
												// noeuds a distance 1 de x
												// (et a distance k-1) plus loin
		this->adjacent_vertices(x, *dk1) ;

		vertices.insert(dk1->begin(), dk1->end()) ;

												// vertices contient les noeuds
												// a distance au plus 1 de x
		if (n==1)
		{
			delete dk1 ;
			return ;
		}
												// n >= 2
												// on iterera pour k = 2 a n

		Tvertice_identifiers_set dmaxk2 ;		// noeuds a distance au plus
		dmaxk2.insert(x) ;						// k-2 de x

												// pour chaque distance k
												// de 2 a n
		for (int k = 2 ; k <= n ; k++)
		{
			Tvertice_identifiers_set* dk = new Tvertice_identifiers_set() ;

												// recupere dans dk les noeuds
												// a distance k de x
			this->distance(x, k, *dk, *dk1, dmaxk2) ;

												// les noeuds a distance au plus k
												// sont dans dmaxk2 union dk1 union dk
			if (k < n)
			{
				dmaxk2.insert(dk1->begin(), dk1->end()) ;

				delete dk1 ;
				dk1 = dk   ;
			}
			else
			{
												// les noeuds a distance au plus n
												// sont dmaxk2 union dk1 union dk
				vertices.insert(dmaxk2.begin(), dmaxk2.end()) ;
				vertices.insert(dk1->begin(), dk1->end()) ;
				vertices.insert(dk->begin(), dk->end()) ;

				delete dk1 ;
				delete dk ;
			}
		}
}


// ============================== ABSTRACTIONS NOEUDS ==========================

				// elimine les noeuds isoles
template <typename Tvalue>
bool graph<Tvalue>::discard_isolated_vertices()
{
	bool discarded_vertices = false ;

	Tadjacences_map_iterator it = this->get_adjacences_map().begin() ;

	while (it != this->get_adjacences_map().end())
	{
		if (it->second.predeccessors.empty() && it->second.successors.empty())
		{
											// noeud isole
			const Tvertice_identifier& ident = it-> first ;

			this->get_vertice_identifiers_list().erase(ident) ;
			this->get_vertices_map().erase(ident) ;

			it = this->get_adjacences_map().erase(it) ;

			discarded_vertices = true ;
		}
		else
			it++ ;
	}

	if (this->connected_components_computed)
			this->recompute_connected_components = true ;

	return discarded_vertices ;
}



// conserve les ARETES (x,y) telles que x et y ont au moins k-2 voisins communs
template <typename Tvalue>
bool graph<Tvalue>::abstraction_k_dense(int min_sup, int k)
{
									// temoigne qu'au moins un sommet ou une arete a ete retire
	bool discarded_vertice = false ;
	bool discarded_edge    = false ;

									// recherche d'un point fixe de l'abstraction
    do
	{								// on reduit d'abord le graphe aux noeuds
									// de degre au moins k-1
    								// (x lie a� y, et x possa�de en plus k-2 voisins (idem pour y), donc k-1 voisins)
		discarded_vertice = this->abstraction_minimal_degree(min_sup, k-1) ;
		// this->display_graph() ;

												// pour chaque sommet x du graphe
		if(this->nb_vertices() >= min_sup)
		{
			discarded_edge = false ;

			Tvertice_identifiers_set_iterator it_x = this->vertice_identifiers_set.begin() ;

			for ( ; it_x != this->vertice_identifiers_set.end() ; it_x++)
			{
				// cout << "x = " << this->get_vertice_value(*it_x) << endl ;

				Tvertice_identifiers_set& successors_x = this->get_successors(*it_x) ;
				Tvertice_identifiers_set_iterator it_y = successors_x.begin() ;

																		// pour chaque successeur y de x
				while (it_y != successors_x.end())
				{
																		// on considere l'arete (x,y)

					// cout << "y = " << this->get_vertice_value(*it_y) << endl ;

					if(! sufficient_common_adjacent_vertices(*it_x,*it_y,k-2))
					{													// arete (x,y) non satisfaisante

																		// efface l'arete (x,y)
						this->get_predeccessors(*it_y).erase(*it_x) ;   // efface x des predecesseurs de y
						it_y = successors_x.erase(it_y)    ;  			// efface y des successeurs de x
						discarded_edge = true ;
					}
					else
						it_y++ ;
				}
				// fin pour chaque successeur y de x
			}
			// fin pour chaque sommet x
		}
		// fin if (this->nb_vertices() >= min_sup)
	}
	while(this->nb_vertices() >= min_sup && discarded_edge) ;

	return discarded_vertice ;
}
// fin abstraction_k_dense


	
template <typename Tvalue>
bool graph<Tvalue>::abstraction_near_star_in(float min_degree_in, float max_distance)
{
	return this->abstraction_near_star(min_degree_in, max_distance) ;
}


template <typename Tvalue>
bool graph<Tvalue>::abstraction_near_star_out(float min_degree_out, float max_distance)
{
	return this->abstraction_near_star(min_degree_out, max_distance) ;
}


	// conserve un sommet s'il est de degre superieur ou egale a� min_degre,
	// ou s'il est a� distance au plus max_distance d'un sommet de degre
	// superieur ou egal a� min_degre
template <typename Tvalue>
bool graph<Tvalue>::abstraction_near_star(float min_degre, float max_distance)
{
	Tvertice_identifiers_set vertices_ok ;
	Tvertice_identifiers_set_iterator it = this->vertice_identifiers_set.begin() ;

											// pour chaque sommet du graphe
	while(it != this->vertice_identifiers_set.end())
	{
		// cout << "*it = " << *it << endl ;

		float degree = this->degree(*it) ;

		// cout << "degree = " << degree << endl ;

		if (degree >= min_degre)
		{
			Tvertice_identifiers_set dmaxk ;

			this->distance_max(*it, max_distance, dmaxk) ;

			// displaySet("dmaxk", dmaxk) ;

			vertices_ok.insert(dmaxk.begin(), dmaxk.end()) ;
		}

		it++ ;
	}

	if (vertices_ok.size() < this->nb_vertices())
	{
											// retire les sommets qui ne sont pas ok
		Tvertice_identifiers_set diff ;

		set_difference(
						this->vertice_identifiers_set.begin(), this->vertice_identifiers_set.end(),
						vertices_ok.begin(), vertices_ok.end(),
						inserter(diff, diff.begin())
					   ) ;

		// graph::display_vertice_identifiers_set(diff) ;
		// cout << endl ;

		// displaySet("diff", diff) ;
											 // diff contient les noeuds a� retirer
		this->discard_vertices(diff) ;

		return true ;
	}

return false ;							// aucun sommet retire
}


template <typename Tvalue>
bool graph<Tvalue>::abstraction_classic_hub_autority(int min_sup, float min_hub, float min_autority)
{
	cout << "the abstract classic_hub_autority does not apply on non directed graph !!!!" << endl ;
	return false ;
}


template <typename Tvalue>
bool graph<Tvalue>::abstraction_classic_hub_autority_pred(int min_sup, float min_hub,
														 float min_autority,
														  Tvertice_identifiers_set& H,
														  Tvertice_identifiers_set& A)
{
	cout << "the abstract classic_hub_autority does not apply on non directed graph !!!!" << endl ;
	return false ;
}


/*
	abstraction topologique
 */
template <typename Tvalue>
bool graph<Tvalue>::abstraction_topologique(int min_sup, float alpha, graph<string>* initial_graph)
{
	bool discarded_vertice = false ;
	bool changed = true ;

	while(this->nb_vertices() >= min_sup && changed)
	{
		changed = false ;

		Tvertice_identifiers_set_iterator x = this->vertice_identifiers_set.begin() ; 

		
												// parcours des sommets du graphe
		while (this->nb_vertices() >= min_sup && x != this->vertice_identifiers_set.end())
		{

			float degre_dans_G  = initial_graph->map_degres[*x] + 1 ;
			float degre_dans_GA = this->degree(*x) + 1 ;
			float proportion  = (degre_dans_GA * 1.0) / degre_dans_G ;


			// cout << "*x             = " << *x << endl ;
			// cout << "degre dans G   = " << degre_dans_G << endl ;
			// cout << "degre dans Gp  = " << degre_dans_GA << endl ;
			// cout << "proportion     = " << proportion  << endl ;
			// cout << "alpha          = " << alpha << endl ;

												// verifie si le sommet *x
												// satisfait la condition
			if (proportion >= alpha)
			{									// x est ok
					x++ ;
			}
			else		
			{
													// recupere les noeuds adjacents a� x
				Tvertice_identifiers_set adjacent_vertices_x ;
				this->adjacent_vertices(*x, adjacent_vertices_x) ;
				
				auto y = adjacent_vertices_x.begin() ;
				bool y_ok = false ;
													// parcourt les noeuds adjacents a� x
				while(y != adjacent_vertices_x.end() && !y_ok)
				{
													// y est adjacent a� x
					if ((this->degree(*y) + 1) / (initial_graph->map_degres[*y] +1)  >= alpha)
					{								// y est ok
						y_ok = true ;
					}
					else
						y++ ;						// y suivant
				}
				// fin while sur y
			
				if (!y_ok)
				{
					// cerr << " : discarded " << endl ;
					changed = true ;
														// on supprime (*it) dans toutes les donnees oa� il apparait
														// sauf dans vertices_list
					this->discard_vertice_elements(*x, true, true, false) ;

														// on supprime (*it) de vertices_list
					x = this->vertice_identifiers_set.erase(x) ;

					discarded_vertice = true ;
				}
				else
				{
						// cerr << endl ;
						x++ ;
				}
			}
		} 
		// fin wile sur x
	}

	return discarded_vertice ;
}


/*
 * on retire tous les noeuds (et arcs y faisant reference) qui ne sont
 * pas de degre au moins "degre"
 * SI LES COMPOSANTES CONNEXES ONT ETE CALCULEES,
 * ELLES SONT MISES A JOUR EN CONSEQUENCES
 */
template <typename Tvalue>
bool graph<Tvalue>::abstraction_minimal_degree(int min_sup, float min_degre)
{
	bool discarded_vertice = false ;
	bool changed = true ;

	while(this->nb_vertices() >= min_sup && changed)
	{
		changed = false ;

		Tvertice_identifiers_set_iterator it = this->vertice_identifiers_set.begin() ;

		while (this->nb_vertices() >= min_sup && it != this->vertice_identifiers_set.end())
		{
			// cerr << "*it = " << *it  ;

			if (this->degree(*it) < min_degre)
			{

				// cerr << " : discarded " << endl ;
				changed = true ;
												// on supprime (*it) dans toutes les donnees oa� il apparait
												// sauf dans vertices_list
				this->discard_vertice_elements(*it, true, true, false) ;

												// on supprime (*it) de vertices_list
				it = this->vertice_identifiers_set.erase(it) ;

				discarded_vertice = true ;
			}
			else
			{
				// cerr << endl ;
				it++ ;
			}

			// cerr << "nb noeudes = " << this->nb_vertices() << endl ;
		}
	}

	return discarded_vertice ;
}


/*
	 on se ramene a l'abstraction de degre minimale d'un graphe non oriente
	 * SI LES COMPOSANTES CONNEXES ONT ETE CALCULEES,
	 * ELLES SONT MISES a� JOUR EN CONSEQUENCES
	 */
	template <typename Tvalue>
	bool graph<Tvalue>::abstraction_minimal_degree_in_out(int min_sup, float min_in, float min_out)
	{
		return this->abstraction_minimal_degree(min_sup, min_in + min_out) ;
	}


	//			voir si on peut utiliser les lambdas ou pointeur de fonction
		//			pour avoir une methode generique (valable pour plusieurs fonction
		/*
		 * on retire tous les noeuds (et arcs y faisant reference) qui ne sont
		 * pas de degre entrant au moins in, et de degre sortant au moint out
		 * SI LES COMPOSANTES CONNEXES ONT ETE CALCULEES,
		 * ELLES SONT MISES A JOUR EN CONSEQUENCES
		 */
	template <typename Tvalue>
	bool graph<Tvalue>::abstraction_minimal_degre_simple(int min_sup, float min_degre)
	{
		return this->abstraction_minimal_degree(min_sup, min_degre) ;
	}


/*
 * on retire tous les noeuds qui ne sont pas dans une composante connexe
 * de taille au moins k
 * les composantes connexes de taille < k sont eliminees
 * => EN SORTIE LES COMPOSANTES CONNEXES SONT CORRECTES
 */
template <typename Tvalue>
bool graph<Tvalue>::abstraction_minimal_connected_components(int min_sup, int min_cc)
{
	bool discarded_vertice = false ;

	// this->display_graph() ;

	if (this->recompute_connected_components)
		this->update_connected_components() ;

	if (!this->connected_components_computed)
		this->compute_connected_components() ;

	// this->display_graph() ;

	Tconnected_components_iterator it = this->get_connected_components().begin() ;

	while (this->nb_vertices() >= min_sup && it != this->get_connected_components().end())
	{
		// this->display_vertice_identifiers() ;
		// cout << "nb_vertices = " << this->nb_vertices() << endl ;

		// cout << "num composante = " << it->first << endl ;
		// cout << "taille composante = " << it->second.size() << endl ;

		// cout << "composante = " ;
		// graph::display_vertice_identifiers_set(it->second) ;
		// cout << endl ;
										// *it est une composante connexe
		if ((it->second).size() < min_cc)
		{
			this->discard_vertices(it->second) ;
			discarded_vertice = true ;
									// ON A DETRUIT LES NOEUDS, PAS LES
									// LA COMPOSANTE
									// OPTION recompute_connected_composante = true
		}
		it++ ;
	}

	return discarded_vertice ;
}
// fin abstractionComposanteConnexe


	// on retire du graphe tous les noeuds qui ne sont pas dans un triangle
template <typename Tvalue>
bool graph<Tvalue>::abstraction_triangles(int min_sup)
{

	Tvertice_identifiers_set apexs_triangles ;
	Tadjacences_map_iterator it_end  = this->get_adjacences_map().end() ;

												// pour chaque noeud du graphe
												// (adjacences_map contient
												// <id_noeud, {id_predecesseurs}, {id_successeurs}>
	for (Tadjacences_map_iterator it = this->adjacences_map.begin() ; it != it_end ; it++)
	{
		Tvertice_identifier vertice = it->first ;

		// displaySet("apexs", apexs_triangles) ;

		// cout << "vertice = " << vertice << endl ;

		if (apexs_triangles.find(vertice) == apexs_triangles.end())
		{										//vertice pas deja� vu dans un triangle

			Tvertice_identifiers_set& successors   = it->second.successors ;
			Tvertice_identifiers_set& predeccessors = it->second.predeccessors ;

			// displaySet("sucesseurs", successors) ;
			// displaySet("predecesseurs", predeccessors) ;

			if (successors.size() + predeccessors.size() >= 2)
			{									// pour chaque couple de predecesseurs union predeccesseurs
												// jusqu'a� triangle trouve

				Tvertice_identifiers_set linked_nodes ;

				set_union(predeccessors.begin(), predeccessors.end(),
				          successors.begin(), successors.end(),
				          inserter(linked_nodes, linked_nodes.begin()));

				// displaySet("linked nodes = ", linked_nodes) ;

				Tvertice_identifiers_set_iterator it2_end = linked_nodes.end() ; // fin list des noeuds lies
				Tvertice_identifiers_set_iterator it1     = linked_nodes.begin() ;
				Tvertice_identifiers_set_iterator it1_end = prev(it2_end,1)  ; // avant dernier successeurs

				bool triangle = false ;

				while (!triangle && it1 != it1_end)
				{
					Tvertice_identifiers_set_iterator it2 = next(it1,1) ;
												// pour chaque couple (*it1, *it2) des noeuds lies a� vertice
												// on verifie que *it1, *it2 sont lies

					while(!triangle && it2 != it2_end)
					{
												// on verifie que *it.first, *it1 et *it2 forme un triangle

												// recherche les noeuds lies a� *it1
						Tvertice_identifiers_set& successors1    = this->get_successors(*it1) ;
						Tvertice_identifiers_set& predeccessors1 = this->get_predeccessors(*it1) ;

												// recherche si *it2 est lie a� *it1
						bool b1 = successors1.find(*it2)   != successors1.end() ;
						bool b2 = predeccessors1.find(*it2) != predeccessors1.end() ;

						if (b1 || b2)
						{
												// *it1 est lie a� *it2
							triangle = true ;

												// ajout des 3 sommets comme sommets d'un triangle
							apexs_triangles.insert(vertice) ;
							apexs_triangles.insert(*it1) ;
							apexs_triangles.insert(*it2) ;
						}
						else
							it2++ ;
					}
					// fin for it2

					if (!triangle)
						it1++ ;
				}
				// fin for it1
		}
		// fin if successors.size() > 2

	  }
		// fin if apexs_triangles.find(vertice) == apexs_triangles.end())
	}
	// fin for it

	// cout << "nombre vertices = " << this->nb_vertices() << endl ;
	// cout << "nb noeuds triangles = " << apexs_triangles.size() << endl ;

	return this->reduce_graph(apexs_triangles) ;
}


	// return the number of edges x-y or y-x such that x is in the current graph and
	// y is not (y is in the initial_graph that includes the current graph)
template <typename Tvalue>
int graph<Tvalue>::
compute_mcb(
			graph<string>* c_graph,   // graph precedent dans la specialisation
		    map<Tvertice_identifier, Tvertice_identifiers_set>& c_edges // arc valides graph precedent
			)
{
		// sommets du graphe courant
	Tvertice_identifiers_set& ve = this->get_vertice_identifiers_list() ;

		// sommets du graphe precedent
	Tvertice_identifiers_set& v  = c_graph->get_vertice_identifiers_list() ;

	// displaySet("ve", ve) ;
    // displaySet("v", v) ;

    // displayMapSet("edges avant retrait", c_edges) ;
    // cout << endl ;

	map<Tvertice_identifier, Tvertice_identifiers_set>::iterator ite ;
									// on retire les arc de c_edges qui ne sont plus valides
	ite = c_edges.begin() ;

	while(ite != c_edges.end())
	{
		Tvertice_identifier x = ite->first ;

		// cout << "x = " << x << endl ;
		// displaySet("successeurs ", ite->second) ;

		if (ve.find(x)== ve.end())
		{
			// cout << "retire " << endl ;
									// x n'est pas dans dans le graphe courant
									// on l'elimine de c_edges
			ite = c_edges.erase(ite) ;
		}
		else
			ite++ ;
	}

	// displayMapSet("edges avant ajout", c_edges) ;
	// cout << endl ;

	// cout << "ajouts..." << endl ;
									// on ajoute des nouveaux arcs :
									// pour chaque sommet x du graphe courant
									// on considere les arretes x-y ou y-x tel que
									// y est dans c_graph et pas dans le graphe courant
	for (Tvertice_identifiers_set_iterator it = ve.begin() ; it != ve.end() ; it++)
	{
		Tvertice_identifier x = *it ;

		// cout << "noeud : " << x << endl ;
													// on recupere dans c_graph
													// les predecesseurs et successeurs de x
		Tvertice_identifiers_set& pred = (c_graph->adjacences_map)[x].predeccessors ;
		Tvertice_identifiers_set& succ = (c_graph->adjacences_map)[x].successors ;

		// displaySet("pred", pred) ;
		// displaySet("succ", succ) ;

		Tvertice_identifiers_set hors_graph ;
													// on identifie les predecesseurs de x
													// hors du graphe courant
		set_difference(pred.begin(), pred.end(), ve.begin(), 
		               ve.end(), inserter(hors_graph, hors_graph.begin()));

		// displaySet("hors_graph1", hors_graph) ;

													// on y ajoute les successeurs de x
													// hors du graphe courant
		set_difference(succ.begin(), succ.end(), ve.begin(), ve.end(), 
		               inserter(hors_graph, hors_graph.begin()));

		// displaySet("hors_graph2", hors_graph) ;

		// cout << "hors_graph.size() = " << hors_graph.size() << endl ;

												// hors graph contient les noeuds y
												// tels que x-y existe et y hors du graphe
												// courant
		if (!hors_graph.empty())
		{
			// cout << "ajoute" << endl ;

			ite = c_edges.find(x) ;
			if (ite == c_edges.end())
			{
				c_edges.insert(pair<Tvertice_identifier,Tvertice_identifiers_set>(x,hors_graph));
			}
			else
			{
				  Tvertice_identifiers_set temp ;

				  set_union((ite->second).begin(), (ite->second).end(),
				            hors_graph.begin(), hors_graph.end(),
				            inserter(temp, temp.begin()));

				  ite->second = temp ;

				  // displaySet("ite->second", ite->second) ;
			}
		}

		// cout << "mcb = " << mcb << endl ;
	}

	// cout << "passe "<< endl ;

	// displayMapSet( "edges final", c_edges) ;
	// cout << endl ;

		// retourne le nombre d'aretes
	int nb_edges = 0 ;

	for (ite = c_edges.begin() ; ite != c_edges.end() ; ite++)
	{
		nb_edges += (ite->second).size() ;
	}

    return nb_edges ;
}


	// check wether the graph satisfies the constraint of local modularity
	// and minimal frequency
	// return true if we continue the specialization process
	// keep_pattern == true if the pattern must be output
	// (it is possible that we return true and output_patter is false)
template <typename Tvalue>
bool graph<Tvalue>::
graph_ok(int min_sup,
		 statsModularity& statsMod,
		 bool modularity_on,
		 float modularity,
		 graph<string>* initial_graph,
		 graph<string>* c_graph,			// graph precedent dans la specialisation
		 map<Tvertice_identifier, Tvertice_identifiers_set>& c_edges,	// arcs (special mod) du graphe precedent
		 bool& output_pattern)
{
	// int nb_vertices = this->nb_vertices() ;
	// cout << "nb vertices = " << nb_vertices << endl ;

	bool cont = this->nb_vertices() >= min_sup ;

	if (cont)
	{
	    if (modularity_on)
	    {
		   return modularity_ok(statsMod, modularity,
				   	   	        initial_graph, c_graph, c_edges, output_pattern) ;
		}
		else
		{
		    output_pattern = true ;
		    return true ;
		}    
	}

    output_pattern = false ;
	return false ;
}


	// verification de la modularite locale
template <typename Tvalue>
bool graph<Tvalue>::
modularity_ok(
			  statsModularity& statsMod,
			  float modularity,
			  graph<string>* initial_graph,
			  graph<string>* c_graph,
			  map<Tvertice_identifier, Tvertice_identifiers_set>& c_edges,
              bool& output_pattern)
{
												// pas int  pour question de calcul
		float m    = initial_graph->nb_edges() ;
		float mc   = this->nb_edges() ;
		float mc_m = mc/m ;
		float mod_star ;

													// modularite locale estimee
		if (mc >= m/2)
			mod_star = 0.25 ;
		else
			mod_star = mc_m * (1 - mc_m) ;

		// cout << "mod* = " << mod_star << endl ;

		if (mod_star < modularity)
		{
		    output_pattern = false ;
			return false ;		// on elague
	    }		

		statsMod.incNbLme() ;

		// cout << "lme = " << statsMod.getNbLme() << endl ;

				// calcul modularite locale

		int mcb = this->compute_mcb(c_graph, c_edges) ;	// mcb = #aretes ayant une
		                                        		// extremite dans e (this) et
		                                        		// l'autre hors (this)
														// et l'autre hors de e

		// cout << "mcb = " << mcb << endl ;

		float mod = mc_m - (1/(4*m*m))*(2*mc + mcb) * (2*mc + mcb) ;

												// le pattern n'est considere que
												// si mod >= modularity
		// cout << "mod = " << mod << endl ;

		if (mod >= modularity)
		{
			statsMod.incNbLm() ;
			// cout << "lm = " << statsMod.getNbLm() << endl ;

			output_pattern = true ;
		}
		else
			output_pattern = false ;

		// cout << "output = " << output_pattern << endl ;

		return true ;			                // pas elagage, possible output_pattern = false
}


	// applique l'ensemble des abstractions demandees
template <typename Tvalue>
void graph<Tvalue>::
apply_abstractions(abstraction& a, int min_sup, graph<string>* initial_graph, Troles_set& roles)
{
	bool changed_near_star     ;
	bool changed_near_star_in  ;
	bool changed_near_star_out ;
	bool changed_triangles     ;
	bool changed_k_dense       ;
	bool changed_degree_in_out ;
	bool changed_degre_simple  ;
	bool changed_abstraction_topologique  ;
	bool changed_classic_hub_autority ;

					// we apply the abstractions until a fixed point is reached
	do
	{
		changed_k_dense        		 	 = false ;
		changed_near_star      		 	 = false ;
		changed_near_star_in		 	 = false ;
		changed_near_star_out		 	 = false ;
		changed_triangles      		 	 = false ;
		changed_degree_in_out 		 	 = false ;
		changed_degre_simple   		 	 = false ;
		changed_abstraction_topologique  = false ;
		changed_classic_hub_autority     = false ;
		
		// cout << "nb vertices = " << this->nb_vertices() << endl ;
		// cout << "nb edges = "<< this->nb_edges() << endl ;

		// this->save_graph_dot("avantabstraction2", true, true) ;

		if (this->nb_vertices() >= min_sup && a.get_degree() > 0)
			this->abstraction_minimal_degree(min_sup, a.get_degree()) ;

		// cout << "nb vertices = " << this->nb_vertices() << endl ;
		// cout << "nb edges = "<< this->nb_edges() << endl ;

		// this->save_graph_dot("apresabstraction2", true, true) ;

		if (this->nb_vertices() >= min_sup && a.get_degre_simple() > 0)
			changed_degre_simple = this->abstraction_minimal_degre_simple(min_sup, a.get_degre_simple()) ;

		if (this->nb_vertices() >= min_sup && (a.get_degree_in() > 0 || a.get_degree_out() >0))
			changed_degree_in_out = this->abstraction_minimal_degree_in_out(min_sup, a.get_degree_in(),
									a.get_degree_out()) ;

		// cout << this->nb_vertices() << endl ;

		if (this->nb_vertices() >= min_sup && a.get_k_dense() > 0)
		{
			// cerr << "k-dense..." << endl ;
			changed_k_dense = this->abstraction_k_dense(min_sup, a.get_k_dense()) ;
			// cerr << "end k-dense." << endl ;
		}

		if (this->nb_vertices() >= min_sup && a.get_classic_hub_autority().first > 0)
		{
			Tvertice_identifiers_set H ;
			Tvertice_identifiers_set A ;

			roles.erase(HUB) ;
			roles.erase(AUTORITE) ;
			changed_classic_hub_autority = this->abstraction_classic_hub_autority_pred(min_sup,
							                             a.get_classic_hub_autority().first,
									                     a.get_classic_hub_autority().second,
														 H, A) ;
			roles[HUB]      = H ;
			roles[AUTORITE] = A ;
		}


		if (this->nb_vertices() >= min_sup && a.get_near_star().first > 0)
			changed_near_star = this->abstraction_near_star(a.get_near_star().first,
					                                        a.get_near_star().second) ;

		if (this->nb_vertices() >= min_sup && a.get_near_star_in().first > 0)
			changed_near_star_in = this->abstraction_near_star_in(a.get_near_star_in().first,
					                                              a.get_near_star_in().second) ;

		if (this->nb_vertices() >= min_sup && a.get_near_star_out().first > 0)
			changed_near_star_out = this->abstraction_near_star_out(a.get_near_star_out().first,
					                                                a.get_near_star_out().second) ;

		if (this->nb_vertices() >= min_sup && a.get_triangles())
			changed_triangles = this->abstraction_triangles(min_sup) ;

		if (this->nb_vertices() >= min_sup && a.get_abstraction_topologique() > 0)
			changed_abstraction_topologique = this->abstraction_topologique(min_sup, a.get_abstraction_topologique(), initial_graph) ;
	}
	while ((changed_degree_in_out || changed_k_dense || changed_near_star || changed_degre_simple ||
			changed_classic_hub_autority || changed_near_star_in || changed_near_star_out ||
		 	changed_triangles || changed_abstraction_topologique) &&
			this->nb_vertices() >= min_sup) ;

					// on abstrait apra�s eventuellement sur les composantes connexes

	if (this->nb_vertices() >= min_sup && a.get_connected_components() > 0)
	{   // cerr << "connected components..." << endl ;
		this->abstraction_minimal_connected_components(min_sup, a.get_connected_components()) ;
		// cerr << "end connected components." << endl ;
	}
}

	// ------------------- CLIQUES-------------------------------------

	// retourne true s'il existe une k-clique dans vertices, retourne une des k-cliques
template <typename Tvalue>
bool graph<Tvalue>::existe_k_clique(
		const int k,												// taille de la k-clique
		Tvertice_identifiers_set& vertices,							// ensemble de noeuds oa� chercher la k-clique
		Tvertice_identifiers_set& clique)							// clique retournee (valide si true)
{

	if (vertices.size() < k)
		return false ;

	if (k==0)
		return true ;					// clique vide

	return check_k_clique(k, vertices, clique) ;

}



// retourne true s'il existe une k-clique dans vertices
template <typename Tvalue>
bool graph<Tvalue>::check_k_clique(
		const int k,												// nombre de noeuds a� ajouter a� clique pour la completer
																	// (k >= 1)
		Tvertice_identifiers_set& vertices,							// noeuds pouvant a�tre ajoutes a� clique (ils sont tous
																	// adjacents a� tous les noeuds de clique)
																	// on a vertices.size() >= k
		Tvertice_identifiers_set& clique)							// clique en construction
{
									// vertices.size() >= k

	if (k==1)						// on peut prendre le 1er element de vertices
	{
		clique.insert(*vertices.begin()) ;
		return true ;
	}

									// k >= 2
	bool found_clique = false ;

	Tvertice_identifiers_set_iterator x = vertices.begin() ;

										// parcourt les noeuds possibles pour ajouter a� clique
	while (!found_clique && x != vertices.end())
	{
										// recupere les noeuds adjacents a� *x
		Tvertice_identifiers_set adjacent_vertices_x ;
		this->adjacent_vertices(*x, adjacent_vertices_x) ;

										// calcul les noeuds adjacents a� clique union {x}
		Tvertice_identifiers_set vertices1 ;
		set_intersection(
							vertices.begin(),vertices.end(),
							adjacent_vertices_x.begin(), adjacent_vertices_x.end(),
							inserter(vertices1, vertices1.begin())
						) ;

		if (vertices1.size() >= k-1)
		{
			Tvertice_identifier v = *x ;

			clique.insert(v) ;

			found_clique = this->check_k_clique(k-1, vertices1, clique) ;

			if (!found_clique)
			{							// on restaure les valeurs modifiees
				vertices.insert(v) ;	// x pointe toujours vers l'element suivant v
				clique.erase(v) ;
			}
		}
		else							// pas de clique possible avec *x
			x++ ;
	}
	return found_clique ;
}


#endif /* GRAPH_HPP_ */
