/* version minerLC */

/*
 * weighted_directed_graphe.hpp
 *
 *  Created on: 2019
 *      Author: Dominique Bouthinon
 *
 *  Representation d'un graphe oriente par une liste d'adjacences
 *  les valeurs des noeuds sont de n'importe quel type
 *  represente par le type template Tvalue
 *
 *  - ajout fonctions din et dout
 *
 */

#ifndef WEIGHTED_DIRECTED_GRAPH_HPP_
#define WEIGHTED_DIRECTED_GRAPH_HPP_

#include <functional>
#include <iostream>
#include <algorithm>
#include <exception>
#include <stack>

#include "graph.hpp"
#include "types_graph.hpp"
#include "abstraction.hpp"
#include "utils_divers.hpp"

using namespace std ;

/* =============================== class graph ============================== */

template <typename Tvalue>
class weighted_directed_graph : public weighted_graphe<Tvalue>,
								public directed_graph<Tvalue>
{
	private :

			// somme les poids des aretes vers les noeuds
			// adjacents a ident
	float somme_poids(Tvertice_identifiers_set& vertices, Tvertice_identifier ident, bool in) ;

	public :

	// --------------------------- CONSTRUCTEURS -------------------------------


	weighted_directed_graph() ;				// graph sans information

			 	 	 	 	// construit le graphe avec un nom et une fonction d'affichage
	weighted_directed_graph(
					string aName,
					function<void (Tvalue)> display_vertice_value_function
		 	 	  ) ;


					 // construit le graphe a partir d'une liste de noeuds et
					 // d'adjacences (avec predecesseurs et successeurs).
					 // ATTENTION : les objets internes sont des copies des parametres
	weighted_directed_graph(
					string aName,
					Tvertices_map<Tvalue>& vertices,
					Tadjacences_map& adjacences,
					Tweighted_edges_map& edges_map,
					function<void (Tvalue)> display_vertice_value_function
				   ) ;


	weighted_directed_graph(
					string aName,
					Tvertices_map<Tvalue>& vertices,
					Tadjacences_map& adjacences,
					Tweighted_edges_map& edges_map,
					function<void (Tvalue)> display_vertice_value_function,
					function<string (Tvalue)> toString_value_function
					) ;


	weighted_directed_graph(
					string aName,
					Tvertices_map<Tvalue>& vertices,
					function<void (Tvalue)> display_vertice_value_function
		 	 	 ) ;

					// idem a partir d'un graph g
	weighted_directed_graph(weighted_directed_graph<Tvalue>& g) ;

	virtual ~weighted_directed_graph() ;

	graph<Tvalue>* copy_graph() ;

	float degree_out(Tvertice_identifier ident) ;
	float degree_in(Tvertice_identifier ident) ;

	float dout(Tvertice_identifier x, Tvertice_identifiers_set& A) ;
	float din(Tvertice_identifier x, Tvertice_identifiers_set& H) ;

	void save_graph_dot(string name, bool vertice_ident, bool vertice_label) ;

} ;
// fin classe weighted_directed_graph<Tvalue>

/* ===================== IMPLEMENTATIONS FONCTIONS ========================== */

//------------------------------------ CONSTRUCTEURS ---------------------------

template <typename Tvalue>
weighted_directed_graph<Tvalue>::
weighted_directed_graph() : graph<Tvalue>()
{

}

		 	 	 	 	// construit le graphe avec un nom et une fonction d'affichage
template <typename Tvalue>
weighted_directed_graph<Tvalue>::
weighted_directed_graph(
				string aName,
				function<void (Tvalue)> display_vertice_value_function
	 	 	 ) :
			 graph<Tvalue>(aName, display_vertice_value_function)
{
}
				 // construit le graphe a partir d'une liste de noeuds et
				 // d'adjacences (avec predecesseurs et successeurs).
				 // ATTENTION : les objets internes sont des copies des paramètres

template <typename Tvalue>
weighted_directed_graph<Tvalue>::
weighted_directed_graph(
				string aName,
				Tvertices_map<Tvalue>& vertices,
				Tadjacences_map& adjacences,
				Tweighted_edges_map& edges_map,
				function<void (Tvalue)> display_vertice_value_function
			   ) : graph<Tvalue>(aName, vertices, adjacences, display_vertice_value_function)
{
	this->set_edges_map(edges_map) ;
}

// construit le graphe a partir d'une liste de noeuds et
// d'adjacences (avec predecesseurs et successeurs).
// ATTENTION : les objets internes sont des copies des paramètres
template <typename Tvalue>
weighted_directed_graph<Tvalue>::
weighted_directed_graph(
				string aName,
				Tvertices_map<Tvalue>& vertices,
				Tadjacences_map& adjacences,
				Tweighted_edges_map& edges_map,
				function<void (Tvalue)> display_vertice_value_function,
				function<string (Tvalue)> toString_value_function
				) : graph<Tvalue>(aName, vertices, adjacences,
						          display_vertice_value_function,
								  toString_value_function)
{
	this->set_edges_map(edges_map) ; // copie
}


template <typename Tvalue>
weighted_directed_graph<Tvalue>::
weighted_directed_graph(
				string aName,
				Tvertices_map<Tvalue>& vertices,
				function<void (Tvalue)> display_vertice_value_function
	 	 	 ) : graph<Tvalue>(aName, vertices, display_vertice_value_function)
{
}

				// constructeur par copie
template <typename Tvalue>
weighted_directed_graph<Tvalue>::
weighted_directed_graph(weighted_directed_graph<Tvalue>& g) : graph<Tvalue>(g)
{
	this->set_edges_map(g.get_edges_map()) ;
}

				// DESTRUCTEUR
template <typename Tvalue>
weighted_directed_graph<Tvalue>::
~weighted_directed_graph()
{
}

template <typename Tvalue>
graph<Tvalue>*  weighted_directed_graph<Tvalue>::
copy_graph()
{
	return new weighted_directed_graph<Tvalue>(*this) ;
}

//---------------------------- NOEUD----------------------------------------

			// somme les poids des arcs depuis (in == true) ou vers (in == false)
			// les noeuds adjacents a ident
template <typename Tvalue>
float weighted_directed_graph<Tvalue>::somme_poids(Tvertice_identifiers_set& vertices,
		                                           Tvertice_identifier ident,
												   bool in)
{
	float degre = 0 ;

	for (Tvertice_identifiers_set_iterator it = vertices.begin() ;
		 it != vertices.end() ; it++)
	{
		pair<Tvertice_identifier, Tvertice_identifier> p ;

		if (in)
			p = make_pair(*it, ident) ;
		else
			p = make_pair(ident, *it) ;

		degre += (this->get_edges_map()).at(p) ;
	}

	return degre ;
}


// retourne le nombre de laisons de x vers A
template <typename Tvalue>
float weighted_directed_graph<Tvalue>::dout(Tvertice_identifier x, Tvertice_identifiers_set& A)
{
	Tvertice_identifiers_set succ ;
	Tvertice_identifiers_set to_A ;

	this->adjacent_vertices_out(x, succ) ;	// noeuds successeurs de x

	setIntersection(&to_A, succ, A); 		// noeuds de X successeurs de x

					// on somme les poids des arcs x->sucesseur
	float degre = this->somme_poids(to_A, x, false) ;
	return degre ;
}


	// retourne le nombre de laisons de H vers x
template <typename Tvalue>
float weighted_directed_graph<Tvalue>::din(Tvertice_identifier x, Tvertice_identifiers_set& H)
{
	Tvertice_identifiers_set pred   ;
	Tvertice_identifiers_set from_H ;

		this->adjacent_vertices_in(x, pred) ;	// noeuds predecesseurs de x

		setIntersection(&from_H, pred, H);  	// noeuds de X predecesseurs de x

			// on somme les poids des arcs predecesseur->x
		float degre = this->somme_poids(from_H, x, true) ;
		return degre ;
}


/**
 * retourne le degre entrant du noeud vertice (s'il existe)
 * sinon retoune -1
 */
template <typename Tvalue>
float weighted_directed_graph<Tvalue>::degree_in(Tvertice_identifier ident)
{
	if (!this->in_graph(ident))
		return -1 ;

	// recupere les predecesseurs de ident
	Tvertice_identifiers_set& vertices = this->get_predeccessors(ident) ;

	float degre = this->somme_poids(vertices, ident, true) ;
	return degre ;
}


/**
 * retourne le degre sortant du noeud vertice (s'il existe)
 * sinon retoune -1
 */
template <typename Tvalue>
float weighted_directed_graph<Tvalue>::degree_out(Tvertice_identifier ident)
{
	if (!this->in_graph(ident))
		return -1 ;

							// recupere les successeurs de ident
	Tvertice_identifiers_set& vertices = this->get_successors(ident) ;


	float degre = this->somme_poids(vertices, ident, false) ;
	return degre ;
}


// sauvegarde le graph en format dot
template <typename Tvalue>
void weighted_directed_graph<Tvalue>::save_graph_dot(string name, bool vertice_ident, bool vertice_label)
{
	this->save_graph_dot2(name, vertice_ident, vertice_label, true) ;
}



#endif /* weighted_directed_graph_HPP_ */


